''' Class for defining linear model 

Will have an A, B, C, and D matrix for every operating point
u_op, y_op, and x_op

'''

import os
import numpy as np
import scipy as sp
import control as co
import matplotlib.pyplot as plt
import multiprocessing as mp
from itertools import chain
from scipy.io import loadmat

try:
    import pyFAST.linearization.mbc.mbc3 as mbc
except ImportError:
    import weis.control.mbc.mbc3 as mbc
except ImportError:
    raise ImportError('Unable to load mbc3 from pyFAST or WEIS')

class LinearTurbineModel(object):

    def __init__(self, lin_file_dir, lin_file_names, nlin=12, reduceStates=False, fromMat=False, lin_file=None, rm_hydro=False, load_parallel=False):
        '''
            inputs:    
                lin_file_dir (string) - directory of linear file outputs from OpenFAST
                lin_file_names (list of strings) - name of output file from OpenFAST, without extension
                nlin (integer) - number of linearization points
                reduceStates (bool) - State space reduction currently unsupported N/A
                fromMat (bool) - load from .mat file
                lin_file (string) -  name of .mat file from matlab version of mbc3, only used if fromMat=True
                rm_hydro (bool) - remove hydrodynamic states
                load_parallel (bool) - run mbc3 usying multiprocessing
        '''
        if not fromMat:

            # Number of linearization cases or OpenFAST sims, different from nlin/NLinTimes in OpenFAST
            n_lin_cases = len(lin_file_names)

            u_ops = np.array([], [])
            all_MBC = []
            all_matData = []
            all_FAST_linData = []

            if load_parallel:
                import time
                t1 = time.time()
                all_linfiles = [[os.path.realpath(os.path.join(
                    lin_file_dir, lin_file_names[iCase] + '.{}.lin'.format(i_lin+1))) for i_lin in range(0, nlin)] for iCase in range(0,n_lin_cases)]
                cores = mp.cpu_count()
                pool = mp.Pool(cores)
                all_MBC, all_matData, all_FAST_linData = zip(*pool.map(run_mbc3, all_linfiles))
                pool.close()
                pool.join()
                print('loaded in parallel in {} seconds'.format(time.time()-t1))
            else:
                import time
                t1 = time.time()
                for iCase in range(0, n_lin_cases):
                    lin_files_i = [os.path.realpath(os.path.join(
                        lin_file_dir, lin_file_names[iCase] + '.{}.lin'.format(i_lin+1))) for i_lin in range(0, nlin)]
                    MBC, matData, FAST_linData = run_mbc3(lin_files_i)
                    all_MBC.append(MBC)
                    all_matData.append(matData)
                    all_FAST_linData.append(FAST_linData)
                print('loaded in serial in {} seconds'.format(time.time()-t1))


            for iCase in range(0, n_lin_cases):
                # nlin array of linearization outputs for iCase

                MBC = all_MBC[iCase]
                matData = all_matData[iCase]
                FAST_linData = all_FAST_linData[iCase]

                if not iCase:   # first time through
                    # Initialize operating points, matrices
                    u_ops = np.zeros((matData['NumInputs'], n_lin_cases))
                    y_ops = np.zeros((matData['NumOutputs'], n_lin_cases))
                    x_ops = np.zeros((matData['NumStates'], n_lin_cases))
                    omega = np.zeros((nlin, n_lin_cases))

                # operating points
                # u_ops \in real(n_inps,n_ops)
                u_ops[:, iCase] = np.mean(matData['u_op'], 1)

                # y_ops \in real(n_outs,n_ops)
                y_ops[:, iCase] = np.mean(matData['y_op'], 1)

                # x_ops \in real(n_states,n_ops), note this is un-reduced state space model states, with all hydro states
                x_ops[:, iCase] = np.mean(matData['xop'], 1)

                # store rotor speed in rpm of each linearization, omega \in real(n_linear,n_cases)
                omega[:, iCase] = matData['Omega'] * 60 / (2 * np.pi)

                # keep all inputs and outputs
                indInps = np.arange(0, matData['NumInputs']).reshape(-1, 1)
                indOuts = np.arange(0, matData['NumOutputs']).reshape(-1, 1)
                indStates = np.arange(0, matData['NumStates'])

                if not iCase:
                    A_ops = np.zeros((len(indStates), len(indStates), n_lin_cases))
                    B_ops = np.zeros((len(indStates), len(indInps), n_lin_cases))
                    C_ops = np.zeros((len(indOuts), len(indStates), n_lin_cases))
                    D_ops = np.zeros((len(indOuts), len(indInps), n_lin_cases))
                    WindSpeeds = np.zeros(n_lin_cases)

                # A \in real(n_states,n_states,n_ops)
                A_ops[:, :, iCase] = MBC['AvgA'] 
                # B \in real(n_states,n_inputs,n_ops)
                B_ops[:, :, iCase] = MBC['AvgB'] 
                # C \in real(n_outs,n_states,n_ops)
                C_ops[:, :, iCase] = MBC['AvgC'] 
                # D \in real(n_outs,n_inputs,n_ops)
                D_ops[:, :, iCase] = MBC['AvgD']

                # if reduceStates:
                #     if not iCase:
                #         n_states = np.zeros((n_lin_cases, 1), dtype=int)
                #     P = co.StateSpace(A_ops[:, :, iCase], B_ops[:, :, iCase],
                #                       C_ops[:, :, iCase], D_ops[:, :, iCase], remove_useless=False)

                #     # figure out minimum number of states for the systems at each operating point
                #     P_min = co.minreal(P, tol=1e-7, verbose=False)
                #     n_states[iCase] = P_min.states

            # # Now loop through and reduce number of states to maximum of n_states
            # # This is broken!!  Works fine if reduceStates = False and isn't problematic to have all the extra states...yet
            if reduceStates:
                raise RuntimeError('reduceStates=True is not currently supported.')
            #     for iCase in range(0, n_lin_cases):
            #         if not iCase:
            #             self.A_ops = np.zeros((max(n_states)[0], max(n_states)[0], n_lin_cases))
            #             self.B_ops = np.zeros((max(n_states)[0], len(indInps), n_lin_cases))
            #             self.C_ops = np.zeros((len(indOuts), max(n_states)[0], n_lin_cases))
            #             self.D_ops = np.zeros((len(indOuts), len(indInps), n_lin_cases))

            #         P = co.StateSpace(A_ops[:, :, iCase], B_ops[:, :, iCase],
            #                           C_ops[:, :, iCase], D_ops[:, :, iCase], remove_useless=False)
            #         # I don't know why it's not reducing to max(n_states), must add 2
            #         P_red = co.balred(P, max(n_states)[0], method='matchdc')
            #         # P_red = co.minreal(P,tol=1e-7,verbose=False)         # I don't know why it's not reducing to max(n_states), must add 2
            #         self.A_ops[:, :, iCase] = P_red.A
            #         self.B_ops[:, :, iCase] = P_red.B
            #         self.C_ops[:, :, iCase] = P_red.C
            #         self.D_ops[:, :, iCase] = P_red.D

            else:
                self.A_ops = A_ops
                self.B_ops = B_ops
                self.C_ops = C_ops
                self.D_ops = D_ops
            
            # Convert output RPM to rad/s
            rpm_idx = np.flatnonzero(np.core.defchararray.find(matData['DescOutput'], 'rpm') > -1).tolist()
            self.C_ops[rpm_idx, :, :] = rpm2radps(self.C_ops[rpm_idx, :, :])
            matData['DescOutput'] = [desc.replace('rpm', 'rad/s') for desc in matData['DescOutput']]
            # Convert output deg to rad
            deg_idx = np.flatnonzero(np.core.defchararray.find(matData['DescOutput'], 'deg') > -1).tolist()
            self.C_ops[deg_idx, :, :] = deg2rad(self.C_ops[deg_idx, :, :])
            matData['DescOutput'] = [desc.replace('deg', 'rad') for desc in matData['DescOutput']]


            # Save wind speed as own array since that's what we'll schedule over to start
            self.u_ops = u_ops
            self.u_h = self.u_ops[0]
            self.y_ops = y_ops
            self.x_ops = x_ops

            # Input/Output Indices
            self.ind_fast_inps = indInps.squeeze()
            self.ind_fast_outs = indOuts.squeeze()

            # Input, Output, State Descriptions
            self.DescCntrlInpt  = matData['DescCntrlInpt']
            self.DescStates     = matData['DescStates']
            self.DescOutput     = matData['DescOutput']
            self.StateDerivOrder = matData['StateDerivOrder']
            self.omega_rpm      = omega

            # Other important things
            self.n_lin_cases = n_lin_cases

            # Trim the system
            self.trim_system(rm_azimuth=True, rm_hydro=rm_hydro)


        else:  # from matlab .mat file m
            matDict = loadmat(lin_file)

            # operating points
            # u_ops \in real(n_inps,n_ops)
            u_ops = np.zeros(matDict['u_ops'][0][0].shape)
            for u_op in matDict['u_ops'][0]:
                u_ops = np.concatenate((u_ops, u_op), axis=1)

            self.u_ops = u_ops[:, 1:]

            # y_ops \in real(n_outs,n_ops)
            y_ops = np.zeros(matDict['y_ops'][0][0].shape)
            for y_op in matDict['y_ops'][0]:
                y_ops = np.concatenate((y_ops, y_op), axis=1)

            self.y_ops = y_ops[:, 1:]

            # x_ops \in real(n_states,n_ops), note this is un-reduced state space model states, with all hydro states
            x_ops = np.zeros(matDict['x_ops'][0][0].shape)
            for x_op in matDict['x_ops'][0]:
                x_ops = np.concatenate((x_ops, x_op), axis=1)

            self.x_ops = x_ops[:, 1:]

            # Matrices

            # A \in real(n_states,n_states,n_ops)
            n_states = np.shape(matDict['A'][0][0])[0]
            A_ops = np.zeros((n_states, n_states, 1))
            for A_op in matDict['A'][0]:
                A_ops = np.concatenate((A_ops, np.expand_dims(A_op, 2)), axis=2)

            self.A_ops = A_ops[:, :, 1:]

            # B \in real(n_states,n_inputs,n_ops)
            n_states = np.shape(matDict['B'][0][0])[0]
            n_inputs = np.shape(matDict['B'][0][0])[1]
            B_ops = np.zeros((n_states, n_inputs, 1))
            for B_op in matDict['B'][0]:
                B_ops = np.concatenate((B_ops, np.expand_dims(B_op, 2)), axis=2)

            self.B_ops = B_ops[:, :, 1:]

            # C \in real(n_outs,n_states,n_ops)
            n_states = np.shape(matDict['C'][0][0])[1]
            n_outs = np.shape(matDict['C'][0][0])[0]
            C_ops = np.zeros((n_outs, n_states, 1))
            for C_op in matDict['C'][0]:
                C_ops = np.concatenate((C_ops, np.expand_dims(C_op, 2)), axis=2)

            self.C_ops = C_ops[:, :, 1:]

            # D \in real(n_outs,n_inputs,n_ops)
            n_states = np.shape(matDict['D'][0][0])[1]
            n_outs = np.shape(matDict['D'][0][0])[0]
            D_ops = np.zeros((n_outs, n_inputs, 1))
            for D_op in matDict['D'][0]:
                D_ops = np.concatenate((D_ops, np.expand_dims(D_op, 2)), axis=2)

            self.D_ops = D_ops[:, :, 1:]

            # Save wind speed as own array since that's what we'll schedule over to start
            self.u_h = self.u_ops[0]

            # Input/Output Indices
            self.ind_fast_inps = matDict['indInps'][0] - 1

            self.ind_fast_outs = matDict['indOuts'][0] - 1
            # self.ind_fast_outs       = matDict['indOuts'][0][0] - 1


    def get_plant_op(self, u_rot, reduce_states):
        '''
        Interpolate system matrices using wind speed operating point wind_speed_op = mean(u_rot)

        inputs: u_rot timeseries of wind speeds

        '''

        wind_speed_op = np.mean(u_rot)

        if len(self.u_h) > 1:
            f_A = sp.interpolate.interp1d(self.u_h, self.A_ops)
            f_B = sp.interpolate.interp1d(self.u_h, self.B_ops)
            f_C = sp.interpolate.interp1d(self.u_h, self.C_ops)
            f_D = sp.interpolate.interp1d(self.u_h, self.D_ops)

            f_u = sp.interpolate.interp1d(self.u_h, self.u_ops)
            f_y = sp.interpolate.interp1d(self.u_h, self.y_ops)
            f_x = sp.interpolate.interp1d(self.u_h, self.x_ops)

            if wind_speed_op < self.u_h.min() or wind_speed_op > self.u_h.max():
                print('WARNING: Requested linearization operating point outside of range')
                print('Simulation operating point is ' + str(wind_speed_op) + ' m/s')
                print('Linearization operating points from ' +
                      str(self.u_h.min()) + ' to ' + str(self.u_h.max()) + ' m/s')
                print('Results may not be as expected')

            if wind_speed_op < self.u_h.min():
                wind_speed_op = self.u_h.min()

            elif wind_speed_op > self.u_h.max():
                wind_speed_op = self.u_h.max()

            A = f_A(wind_speed_op)
            B = f_B(wind_speed_op)
            C = f_C(wind_speed_op)
            D = f_D(wind_speed_op)

            u_op = f_u(wind_speed_op)
            x_op = f_x(wind_speed_op)
            y_op = f_y(wind_speed_op)
        else:
            print('WARNING: Only one linearization, at ' + str(self.u_h[0]) + ' m/s')
            print('Simulation operating point is ' + str(wind_speed_op) + 'm/s')
            print('Results may not be as expected')

            # Set linear system to only linearization
            A = self.A_ops[:, :, 0]
            B = self.B_ops[:, :, 0]
            C = self.C_ops[:, :, 0]
            D = self.D_ops[:, :, 0]

            u_op = self.u_ops[:, 0]
            y_op = self.y_ops[:, 0]
            x_op = self.x_ops[:, 0]

        # form state space model  TODO: generalize using linear description strings
        P_op = co.StateSpace(A, B, C, D)
        if reduce_states:
            P_op = co.minreal(P_op, tol=1e-7, verbose=False)
        P_op.InputName = self.DescCntrlInpt
        P_op.OutputName = self.DescOutput

        # operating points dict
        ops = {}
        ops['u'] = u_op[self.ind_fast_inps]
        ops['uh'] = wind_speed_op
        ops['y'] = y_op[self.ind_fast_outs]
        ops['x'] = x_op

        return ops, P_op

    def add_control(self, lin_control_model):
        '''
        Add closed loop control to plant, augmenting P_op with LinearControlModel object

        inputs: lin_control_model   - instance of LinearControlModel object
                ops               - wind speed operating point

        OutputName: P_cl               - linear closed-loop plant

        '''

        # find pitch control operating point
        pitch_op = self.ops['u'][1]
        lin_control_model.pitch_control(pitch_op)

        lin_control_model.connect_elements()
        #    TODO: add torque control

        P_cl = connect_ml([self.P_op, lin_control_model.C_all], self.DescCntrlInpt, self.DescOutput)

        return P_cl

    def solve(self, disturbance, Plot=False, open_loop=True, controller={}, reduce_states=False):
        ''' 
        Run linear simulation of open-loop turbine model
        inputs: disturbance: dict containing
                    Time - vector of time indices
                    Wind - vector of wind speeds (usually rotor avg wind speed)
                Plot - plot solution?
                open_loop - (logical) run linearization in open loop or add closed-loop linear controllers
                controller (optional) - linear controller to add if open_loop = False

        outputs: OutList - list of output channels, mimicking nonlinear OpenFAST
                 OutData - array of output channels, mimicking nonlinear OpenFAST
        '''
        # Unpack disturbance
        tt = disturbance['Time']
        u_h = disturbance['Wind']

        # Get plant operating point
        self.ops, self.P_op = self.get_plant_op(u_h, reduce_states)

        if not open_loop:
            if isinstance(controller, LinearControlModel):
                P_op = self.add_control(controller)
            else:
                print('WARNING: controller not LinearControlModel() object')

        # linearize input (remove wind_speed_op)
        u_lin = np.zeros((len(self.DescCntrlInpt), len(tt)))

        indWind = self.DescCntrlInpt.index(
            'IfW Extended input: horizontal wind speed (steady/uniform wind), m/s')
        u_lin[indWind, :] = u_h - self.ops['uh']

        # linear solve
        _, y_lin, xx = co.forced_response(P_op, T=tt, U=u_lin)

        # Add back in operating points
        # Note that bld pitch was an input operating point that we are moving to the output operating point
        # TODO: designate CONTROL inputs that become outputs in closed-loop simulations
        y_op = self.ops['y']  # TODO: make blade pitch operationg point in radians
        u_op = self.ops['u']

        y = y_op.reshape(-1, 1) + y_lin
        u = u_op[0].reshape(-1, 1) + u_lin

        # plot, for debugging
        if Plot:
            chans = ['RtVAvgxh', 'GenSpeed']
            fig, ax = plt.subplots(len(chans), 1)
            for i, chan in enumerate(chans):
                ind = [chan in out for out in self.DescOutput].index(True)      # channel index
                ax[i].plot(tt, y[ind, :])
                ax[i].grid(True)
                ax[i].set_ylabel(chan)

                if i < len(chans):
                    ax[i].set_xticklabels([])
            plt.show()

        # Shorten output names from linearization output to one like level3 openfast output
        # This depends on how openfast sets up the linearization output names and may break if that is changed
        OutList = [out_name.split()[1][:-1] for out_name in P_op.OutputName]
        OutData_arr = y.T

        # Turn OutData into dict like in ROSCO_toolbox
        OutData = {}
        for i, out_chan in enumerate(OutList):
            OutData[out_chan] = OutData_arr[:, i]

        # Add time to OutData
        OutData['Time'] = tt

        return OutData, OutList, P_op

    def trim_system(self, rm_azimuth=True, rm_hydro=True, desInputs=[], desOutputs=[]):
        '''
        Trim states, inputs, and outputs, of system
        '''

        # Find states to remove
        hd_idx = []
        az_idx = []
        if rm_hydro:
            hd_idx = np.flatnonzero(np.core.defchararray.find(self.DescStates, 'HD') > -1).tolist()
        if rm_azimuth:
            azDesc = 'ED Variable speed generator DOF (internal DOF index = DOF_GeAz), rad'
            az_idx = np.flatnonzero(np.core.defchararray.find(self.DescStates, azDesc) > -1).tolist()

        # Define states to keep
        indStates = np.arange(len(self.DescStates))
        indStates = np.delete(indStates, hd_idx + az_idx).reshape(-1, 1)

        # Find inputs to keep
        indInputs = np.arange(0, len(self.DescCntrlInpt))
        if desInputs:
            inp_idx = []
            for inp in desInputs:
                inp_idx += np.flatnonzero(np.core.defchararray.find(self.DescCntrlInpt, inp) > -1).tolist()
            indInputs = indInputs[inp_idx]
        indInputs = indInputs.reshape(-1,1)

        # Find outputs to keep
        indOutputs = np.arange(0, len(self.DescOutput))
        if desOutputs:
            out_idx = []
            for outp in desOutputs:
                out_idx += np.flatnonzero(np.core.defchararray.find(self.DescOutput, outp) > -1).tolist()
            indOutputs = indOutputs[out_idx]
        indOutputs = indOutputs.reshape(-1,1)

        # Setup new state matrices
        nStates  = int(len(indStates))
        nInputs  = len(indInputs)
        nOutputs = len(indOutputs)
        A_nohd = np.zeros(( nStates, nStates, self.n_lin_cases))
        B_nohd = np.zeros(( nStates, nInputs, self.n_lin_cases))
        C_nohd = np.zeros(( nOutputs, nStates, self.n_lin_cases))
        D_nohd = np.zeros(( nOutputs, nInputs, self.n_lin_cases))

        # Remove states
        for case_idx in range(self.n_lin_cases):
            A_nohd[:,:,case_idx] = self.A_ops[:,:,case_idx][indStates, indStates.T].squeeze()
            if len(indInputs) == 1:
                B_nohd[:,:,case_idx] = self.B_ops[:,:,case_idx][indStates, indInputs.T]
            else:
                B_nohd[:,:,case_idx] = self.B_ops[:,:,case_idx][indStates, indInputs.T].squeeze()
            C_nohd[:,:,case_idx] = self.C_ops[:,:,case_idx][indOutputs, indStates.T].squeeze()
            if len(indInputs) == 1:
                D_nohd[:, :, case_idx] = self.D_ops[:, :, case_idx][indOutputs, indInputs.T]
            else:
                D_nohd[:, :, case_idx] = self.D_ops[:, :, case_idx][indOutputs, indInputs.T].squeeze()

        # Replace old state matrices
        self.A_ops = A_nohd
        self.B_ops = B_nohd
        self.C_ops = C_nohd
        self.D_ops = D_nohd

        # Trim Descriptions
        self.DescStates     = np.array(self.DescStates)[indStates].squeeze().tolist()  
        self.x_ops          = self.x_ops[indStates].squeeze()
        self.StateDerivOrder = np.array(self.StateDerivOrder)[indStates].squeeze().tolist()
        self.DescCntrlInpt  = np.array(self.DescCntrlInpt)[indInputs].squeeze().tolist()
        self.DescOutput     = np.array(self.DescOutput)[indOutputs].squeeze().tolist()


class LinearControlModel(object):
    ''' 
    Linear control models from ROSCO input files or ROSCO_Toolbox controller objects

    Control Models:
        - baseline pitch, function of blade pitch operating point?
        - floating feedback

    TODO:
        - PI torque
    
    '''

    def __init__(self, controller, fromDISCON_IN=False, DISCON_file=[], floating=True):
        # Set parameters here for now
        '''
        controller is object from ROSCO Toolbox
        fromDISON_IN if you want to skip that and read directly from DISCON_file object
        '''

        # Define input/output description strings
        self.DescPitch = 'ED Extended input: collective blade-pitch command, rad'
        self.DescGen = 'ED GenSpeed, (rpm)'
        self.DescTwr = 'ED TwrBsMyt, (kN-m)'
        self.DescAz = 'ED Variable speed generator DOF (internal DOF index = DOF_GeAz), rad'
        self.DescPltPitch = 'ED PtfmPitch, (deg)'
        self.DescNacIMUFA = 'ED NcIMURAys, (deg/s^2)'
        # TODO: add torque control signals

        if not fromDISCON_IN:

            # Pitch control parameters, gain scheduled
            self.PC_GS_angles = controller.pitch_op_pc
            self.PC_GS_KP = controller.pc_gain_schedule.Kp
            self.PC_GS_KI = controller.pc_gain_schedule.Ki

            # Floating Control parameters
            self.Fl_Kp = controller.Kp_float
            Fl_Bw, Fl_Damp = controller.turbine.ptfm_freq, 1.0
            fl_lpf = self.low_pass_filter(Fl_Bw, Fl_Damp)

            # Pitch Actuator parameters
            # self.PC_ActBw           = controller.turbine.pitch_act_bw
            self.PC_ActBw = 100  # 1.5708  # hard code until this is pulled into ROSCO

            # Gen Speed Filter parameters
            F_Gen_Freq = controller.turbine.bld_edgewise_freq * 1/4
            F_Gen_Damp = controller.F_LPFDamping

        else:
            # Will probably want to move this out when we have other methods for setting up controller

            # Pitch control parameters, gain scheduled
            self.PC_GS_angles = DISCON_file['PC_GS_angles']
            self.PC_GS_KP = DISCON_file['PC_GS_KP']
            self.PC_GS_KI = DISCON_file['PC_GS_KI']

            # Floating Control parameters
            self.Fl_Kp = DISCON_file['Fl_Kp']
            Fl_Bw, Fl_Damp = DISCON_file['F_FlCornerFreq']
            fl_lpf = self.low_pass_filter(Fl_Bw, Fl_Damp)

            # Pitch Actuator parameters
            self.PC_ActBw = 100  # DISCON_file['PC_ActBw']

            # Gen Speed Filter parameters
            F_Gen_Freq = DISCON_file['F_LPFCornerFreq']
            F_Gen_Damp = DISCON_file['F_LPFDamping']

        # Floating transfer function
        s = co.TransferFunction.s
        self.C_Fl = - fl_lpf / s * np.mean(self.Fl_Kp) * deg2rad(1)
        self.C_Fl = co.ss(self.C_Fl)
        self.C_Fl.InputName = self.DescNacIMUFA
        self.C_Fl.OutputName = 'Fl_Pitch'

        # pitch actuator model
        pitch_act = self.low_pass_filter(self.PC_ActBw, .707)
        self.pitch_act = co.ss(pitch_act)
        self.pitch_act.InputName = 'PitchCmd'
        self.pitch_act.OutputName = self.DescPitch

        # generator filter model
        self.F_Gen = self.low_pass_filter(F_Gen_Freq, F_Gen_Damp)
        self.F_Gen = co.ss(self.F_Gen)
        self.F_Gen.InputName = self.DescGen
        self.F_Gen.OutputName = 'GenSpeedF'

    def pitch_control(self, pitch_op):
        ''' Linear pitch controller, using pitch_op (rad) to determine PI gains

            output: C_PC - control system object with pi pitch controller
        '''

        f_p = sp.interpolate.interp1d(self.PC_GS_angles, self.PC_GS_KP)
        f_i = sp.interpolate.interp1d(self.PC_GS_angles, self.PC_GS_KI)

        kp = f_p(pitch_op)
        ki = f_i(pitch_op)

        s = co.TransferFunction.s
        self.C_PC = -(kp + ki/s) * rpm2radps(1)
        self.C_PC = co.ss(self.C_PC)
        self.C_PC.InputName = 'GenSpeedF'
        self.C_PC.OutputName = 'PC_Pitch'

    def connect_elements(self):
        '''
        Connect various control elements
        required: C_PC, C_Fl, F_Gen, pitch_act
        '''
        # sum block
        S = self.sum_block(['PC_Pitch', 'Fl_Pitch'], 'PitchCmd')

        # Have to connect everything...

        # control modules
        mods = [self.C_PC, self.C_Fl, S, self.F_Gen, self.pitch_act]
        # mods = [self.C_PC,self.C_Fl,S,self.pitch_act]
        # mods = [self.C_PC]

        inVNames = [self.DescGen, self.DescNacIMUFA]
        outVNames = [self.DescPitch]

        self.C_all = connect_ml(mods, inVNames, outVNames)

    def low_pass_filter(self, omega, zeta, order=2):
        '''
            Returns low pass filter control object
        '''

        if order == 2:
            lpf = co.tf([omega**2], [1, 2*zeta*omega, omega**2])
        elif order > 2:
            [b, a] = sp.signal.butter(order, omega, analog=True)
            lpf = co.tf(b, a)
        else:
            lpf = co.tf(omega, [1, omega])

        return lpf

    def sum_block(self, InputName, OutputName):
        '''
            Similar to sumblk() in matlab

            InputName: list of input signals, corresponding to control object I/O names
            OutputName: OutputName signal, corresponding to some control object I/O name


        '''
        S = co.ss(0, [0]*len(InputName), 0, [1]*len(InputName))      # 2D example

        # S = co.tf(S)
        S.InputName = InputName
        S.OutputName = OutputName
        S.dt = 0

        return S


# helper functions
def run_mbc3(fnames):
    '''
    Helper function to run mbc3
    '''
    print('Loading linearizations from:', ''.join(fnames[0].split('.')[:-2]))
    MBC, matData, FAST_linData = mbc.fx_mbc3(fnames, verbose=False)

    return MBC, matData, FAST_linData

def connect_ml(mods, inputs, outputs):
    ''' 
    Connects models like the matlab function does

    inputs: mods  - list of ss models, should have InputName and OutputName fields
            inputs - list of inputs of connected system
            outputs - list of ouptuts of connected system

    output: sys  - connected ss model

    '''

    # Input checking - all timesteps should be the same as the first one
    if mods[0].dt is None:
        mods[0].dt = 0

    dt_main = mods[0].dt

    for mod in mods:
        # timesteps
        if mod.dt is None:
            mod.dt = 0
        elif mod.dt != dt_main:
            print('Some systems have different dts')

        if not hasattr(mod, 'InputName'):
            print('WARNING: missing InputName')

        if not hasattr(mod, 'OutputName'):
            print('WARNING: missing OutputName')

    # append sequentially
    sys = co.ss([], [], [], [])    # init empty system
    sys.dt = 0

    for mod in mods:
        sys = co.append(sys, mod)

    # Set of input/output names
    InNames = []
    for mod in mods:
        if isinstance(mod.InputName, list):
            for InName in mod.InputName:
                InNames.append(InName)
        else:
            InNames.append(mod.InputName)

    OutNames = []
    for mod in mods:
        if isinstance(mod.OutputName, list):
            for OutName in mod.OutputName:
                OutNames.append(OutName)
        else:
            OutNames.append(mod.OutputName)

    # init interconnection matrix
    Q = np.zeros((len(InNames), 2), dtype=int)

    # assign input list into sequential indices
    Q[:, 0] = np.arange(1, len(InNames)+1)

    # assign output list to input indices
    for input_name in InNames:
        if input_name in OutNames:
            Q[InNames.index(input_name), 1] = OutNames.index(input_name)+1

    # Input/Output indices, TODO: better error catching here
    inV = [InNames.index(in_v_name) + 1 for in_v_name in inputs]
    outV = [OutNames.index(out_v_name) + 1 for out_v_name in outputs]

    # Connect...finally!
    # Q = np.array(([1,0,],[2,5],[3,1])).tolist()
    sys = co.connect(sys, Q, inV, outV)
    sys.InputName = inputs
    sys.OutputName = outputs

    return sys

def deg2rad(d):
    return d * np.pi / 180


def rad2deg(r):
    return r * 180 / np.pi


def rpm2radps(r):
    return r * 2 * np.pi / 60


def radps2rpm(r):
    return r * 60 / 2 / np.pi



import os
from wisdem.inputs import validate_with_defaults

schema_dir = os.path.dirname(os.path.abspath(__file__))

def load_rosco_yaml(finput):
    rosco_schema = os.path.join(schema_dir,'toolbox_schema.yaml')
    return validate_with_defaults(finput, rosco_schema)


if __name__=='__main__':
    fname = '/Users/dzalkind/Tools/ROSCO/Tune_Cases/NREL5MW.yaml'
    new_input = load_rosco_yaml(fname)
    
    print('here')




#!/usr/bin/env python
import numpy as np
import matplotlib.pyplot as plt
import os

# Wind file for sheared 18 m/s wind with 30 degree direction.
# Time	Wind	Wind	Vert.	Horiz.	Vert.	LinV	Gust
#	    Speed	Dir	    Speed	Shear	Shear	Shear	Speed

class HH_WindFile(object):
    '''
    Hub height wind file


    __init__: setup file based on wind_type, t_max, dt
    write:    write file
    
    '''

    def __init__(self,**kwargs):


        self.wind_type  = 'step'   # wind file type: 'step' supported
        self.dt         = 0.05    # wind time step
        self.filename   = self.wind_type + '.wnd'  # Path to fst directory files
        self.T_max      = 600

        # set default wind input (step)
        self.time           = [0,300,300+self.dt,600]
        self.wind_speed     = [10,10,14,14]
        self.wind_dir       = [0] * len(self.time)
        self.vert_speed     = [0] * len(self.time)
        self.horiz_shear    = [0] * len(self.time)
        self.vert_shear     = [0] * len(self.time)
        self.linv_shear     = [0] * len(self.time)
        self.gust_speed     = [0] * len(self.time)

        # Optional population class attributes from key word arguments
        for k, w in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(HH_WindFile, self).__init__()


    def write(self):
        if not os.path.isdir(os.path.dirname(self.filename)):
            os.makedirs(os.path.dirname(self.filename))
        with open(self.filename,'w') as f:
            f.write('!\tTime\tWind Speed\tWind Dir\tVert. Spd.\tHoriz. Shr.\t Vert. Shr.\t LinV. Shr.\tGust Speed\n')
            for t, ws, wd, vs, hs, vs, ls, gs in zip(self.time,self.wind_speed,self.wind_dir,self.vert_speed, \
                self.horiz_shear, self.vert_shear, self.linv_shear, self.gust_speed):
                f.write('{:6.6f}\t{:6.6f}\t{:6.6f}\t{:6.6f}\t{:6.6f}\t{:6.6f}\t{:6.6f}\t{:6.6f}\n'.format(
                    t,ws,wd,vs,hs,vs,ls,gs))


    def plot(self):
        plt.plot(self.time,self.wind_speed)
        plt.show()

class HH_StepFile(HH_WindFile):
    
    def __init__(self):

        self.u_start    = 10
        self.u_end      = 14
        self.t_step     = 300
        self.t_max      = 600
        self.dt         = 0.05

        self.wind_type          = 'step'   # wind file type: 'step' supported
        self.wind_directory     = '.'

        self.update()
       

    def update(self):
        self.filename   = os.path.join(self.wind_directory,'{}_{:.1f}_{:.1f}.wnd'.format(self.wind_type,self.u_start,self.u_end))

        # set default wind input (step)
        self.time           = [0,self.t_step,self.t_step+self.dt,self.t_max]
        self.wind_speed     = [self.u_start,self.u_start,self.u_end,self.u_end]
        self.wind_dir       = [0] * len(self.time)
        self.vert_speed     = [0] * len(self.time)
        self.horiz_shear    = [0] * len(self.time)
        self.vert_shear     = [0] * len(self.time)
        self.linv_shear     = [0] * len(self.time)
        self.gust_speed     = [0] * len(self.time)

# def gen_step_wind(,u_start,u_end,t_step,t_max,dt=0.05)
#     '''
#         Generate wind input with single step in wind
#         No shear, direction or anything else

#         Inputs:
#             - filename: where to write file
#             - u_start: starting wind speed
#             - u_end: ending wind speed
#             - t_step: time of wind step

#     '''

#     # time breakpoints
#     t_bp = [0,t_step,t_step+dt,t_max]
#     u_bp = [u_start,u_start,u_end,u_end]


#     Uinf = 8.0
#     max_wdir_delta = 30.0
#     T = 600.0
#     dt = 0.5

#     t = np.arange(0,T+dt,dt)
#     wdir = max_wdir_delta * np.sin(2*np.pi*t/T)
#     for ti,wdi in zip(t,wdir):
#         print('{:>6.1f}\t{:.1f}\t{:.1f}\t0.0\t0.0\t0.0\t0.0\t0.0'.format(ti,Uinf,wdi))



if __name__ == "__main__":
    hh_step = HH_StepFile()

    hh_step.t_step  = 400
    hh_step.t_max   = 800
    hh_step.u_start = 10
    hh_step.u_end   = 11
    hh_step.update()

    hh_step.write()
    
    print('here')



'''
Reads OpenFAST input files. 

Most of this script is copied from:
https://github.com/OpenFAST/python-toolbox/blob/dev/pyFAST/input_output/fast_input_file.py
on 11/05/2020

'''

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from io import open
from builtins import range
from builtins import str
from future import standard_library
standard_library.install_aliases()
from .file import File, WrongFormatError, BrokenFormatError
import os
import numpy as np
import re
import pandas as pd

from ROSCO_toolbox.utilities import read_DISCON

__all__  = ['FASTInputFile']

TABTYPE_NOT_A_TAB          = 0
TABTYPE_NUM_WITH_HEADER    = 1
TABTYPE_NUM_WITH_HEADERCOM = 2
TABTYPE_NUM_NO_HEADER      = 4
TABTYPE_NUM_BEAMDYN        = 5
TABTYPE_MIX_WITH_HEADER    = 6
TABTYPE_FIL                = 3
TABTYPE_FMT                = 9999 # TODO

# --------------------------------------------------------------------------------}
# --- INPUT FILE 
# --------------------------------------------------------------------------------{
class FASTInputDeck():
    '''
    Read/write an OpenFAST input file deck. Behaves like a dictionary
    '''
    def __init__(self, fst_file):
        self.Fst = FASTInputFile(fst_file)
        self.path = os.path.dirname(os.path.abspath(fst_file))

    def load(self):
        fst_dict = {}
        try:
            fst_dict['ElastoDyn'] = self.load_ElastoDyn()
        except:
            "Error loading {}".format(self.Fst['EDFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['BDBldFile(1)'] = self.load_BDBldFile1()
        except:
            "Error loading {}".format(self.Fst['BDBldFile(1)'].strip('\'').strip('\"'))
        
        try:
            fst_dict['BDBldFile(2)'] = self.load_BDBldFile2()
        except:
            "Error loading {}".format(self.Fst['BDBldFile(2)'].strip('\'').strip('\"'))

        try:
            fst_dict['BDBldFile(3)'] = self.load_BDBldFile3()
        except:
            "Error loading {}".format(self.Fst['BDBldFile(3)'].strip('\'').strip('\"'))
        
        try:
            fst_dict['InflowWind'] = self.load_InflowWind()
        except:
            "Error loading {}".format(self.Fst['InflowFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['AeroDyn'] = self.load_AeroDyn()
        except:
            "Error loading {}".format(self.Fst['AeroFile'].strip('\'').strip('\"'))
       
        try:
            fst_dict['ServoDyn'] = self.load_ServoDyn()
            try:
                fst_dict['DISCON_in'] = self.load_DISCON()
            except:
                "Error loading {}".format(self.ServoDyn['DLL_InFile'].strip('\'').strip('\"'))

        except:
            "Error loading {}".format(self.Fst['ServoFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['HydroDyn'] = self.load_HydroDyn()
        except:
            "Error loading {}".format(self.Fst['HydroFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['SubDyn'] = self.load_SubDyn()
        except:
            "Error loading {}".format(self.Fst['SubFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['MoorDyn'] = self.load_MoorDyn()
        except:
            "Error loading {}".format(self.Fst['MooringFile'].strip('\'').strip('\"'))
        
        try:
            fst_dict['IceDyn'] = self.load_IceDyn()
        except:
            "Error loading {}".format(self.Fst['IceFile'].strip('\'').strip('\"'))
        

        return fst_dict

    def load_ElastoDyn(self):
        ED_fullfile = os.path.join(self.path, self.Fst['EDFile'].strip('\'').strip('\"'))
        self.ElastoDyn = FASTInputFile(ED_fullfile)
        return self.ElastoDyn

    def load_BDBldFile1(self):
        BD1_fullfile = os.path.join(self.path, self.Fst['BDBldFile(1)'].strip('\'').strip('\"'))
        self.BD1_fullfile = FASTInputFile(BD1_fullfile)
        return self.load_BDBldFile1

    def load_BDBldFile2(self):
        BD2_fullfile = os.path.join(self.path, self.Fst['BDBldFile(2)'].strip('\'').strip('\"'))
        self.BD2_fullfile = FASTInputFile(BD2_fullfile)
        return self.load_BDBldFile2

    def load_BDBldFile3(self):
        BD3_fullfile = os.path.join(self.path, self.Fst['BDBldFile(3)'].strip('\'').strip('\"'))
        self.BD3_fullfile = FASTInputFile(BD3_fullfile)
        return self.load_BDBldFile3

    def load_InflowWind(self):
        IF_fullfile = os.path.join(self.path, self.Fst['InflowFile'].strip('\'').strip('\"'))
        self.InflowWind = FASTInputFile(IF_fullfile)
        return self.InflowWind

    def load_AeroDyn(self):
        AF_fullfile = os.path.join(self.path, self.Fst['AeroFile'].strip('\'').strip('\"'))
        self.AeroDyn = FASTInputFile(AF_fullfile)
        return self.AeroDyn

    def load_ServoDyn(self):
        SF_fullfile = os.path.join(self.path, self.Fst['ServoFile'].strip('\'').strip('\"'))
        self.ServoDyn = FASTInputFile(SF_fullfile)
        return self.ServoDyn

    def load_HydroDyn(self):
        HF_fullfile = os.path.join(self.path, self.Fst['HydroFile'].strip('\'').strip('\"'))
        self.HydroDyn = FASTInputFile(HF_fullfile)
        return self.HydroDyn

    def load_SubDyn(self):
        SUBF_fullfile = os.path.join(self.path, self.Fst['SubFile'].strip('\'').strip('\"'))
        self.SubDyn = FASTInputFile(SUBF_fullfile)
        return self.SubDyn

    def load_MoorDyn(self):
        MF_fullfile = os.path.join(self.path, self.Fst['MooringFile'].strip('\'').strip('\"'))
        self.MoorDyn = FASTInputFile(MF_fullfile)
        return self.MoorDyn

    def load_IceDyn(self):
        IceF_fullfile = os.path.join(self.path, self.Fst['IceFile'].strip('\'').strip('\"'))
        self.IceDyn = FASTInputFile(IceF_fullfile)
        return self.IceDyn

    def load_DISCON(self): # (TODO - Move this out of the ROSCO toolbox and into ofTools)
        DISCON_fullfile = os.path.join(self.path, self.ServoDyn['DLL_InFile'].strip('\'').strip('\"'))
        self.DISCON_in = read_DISCON(DISCON_fullfile)
        return self.DISCON_in

# --------------------------------------------------------------------------------}
# --- INPUT FILE 
# --------------------------------------------------------------------------------{
class FASTInputFile(File):
    """ 
    Read/write an OpenFAST input file. The object behaves like a dictionary.

    Main methods
    ------------
    - read, write, toDataFrame, keys

    Main keys
    ---------
    The keys correspond to the keys used in the file. For instance for a .fst file: 'DT','TMax'

    Examples
    --------

        filename = 'AeroDyn.dat'
        f = FASTInputFile(filename)
        f['TwrAero'] = True
        f['AirDens'] = 1.225
        f.write('AeroDyn_Changed.dat')

    """

    @staticmethod
    def defaultExtensions():
        return ['.dat','.fst','.txt','.fstf']

    @staticmethod
    def formatName():
        return 'FAST input file'

    def __init__(self, filename=None, **kwargs):
        super(FASTInputFile, self).__init__(filename=filename,**kwargs)

    def keys(self):
        self.labels = [ d['label'] for d in self.data if not d['isComment'] ]
        return self.labels

    def getID(self,label):
        i=self.getIDSafe(label)
        if i<0:
            raise KeyError('Variable `'+ label+'` not found in FAST file:'+self.filename)
        else:
            return i
    def getIDs(self,label):
        I=[]
        # brute force search
        for i in range(len(self.data)):
            d = self.data[i]
            if d['label'].lower()==label.lower():
                I.append(i)
        if len(I)<0:
            raise KeyError('Variable `'+ label+'` not found in FAST file:'+self.filename)
        else:
            return I

    def getIDSafe(self,label):
        # brute force search
        for i in range(len(self.data)):
            d = self.data[i]
            if d['label'].lower()==label.lower():
                return i
        return -1

    # Making object an iterator
    def __iter__(self):
        self.iCurrent=-1
        self.iMax=len(self.data)-1
        return self

    def __next__(self): # Python 2: def next(self)
        if self.iCurrent > self.iMax:
            raise StopIteration
        else:
            self.iCurrent += 1
            return self.data[self.iCurrent]

    # Making it behave like a dictionary
    def __setitem__(self,key,item):
        I = self.getIDs(key)
        for i in I: 
            self.data[i]['value'] = item

    def __getitem__(self,key):
        i = self.getID(key)
        return self.data[i]['value']

    def __repr__(self):
        s ='Fast input file: {}\n'.format(self.filename)
        return s+'\n'.join(['{:15s}: {}'.format(d['label'],d['value']) for i,d in enumerate(self.data)])

    def addKeyVal(self,key,val,descr=None):
        d=getDict()
        d['label']=key
        d['value']=val
        if descr is not None:
            d['descr']=descr
        self.data.append(d)

    def _read(self):

        # TODO members for  BeamDyn with mutliple key point                                                                                                                                                                                                                                                                                                        ####### TODO PropSetID is Duplicate SubDyn and used in HydroDyn
        NUMTAB_FROM_VAL_DETECT  = ['HtFract'  , 'TwrElev'   , 'BlFract'  , 'Genspd_TLU' , 'BlSpn'        , 'WndSpeed' , 'HvCoefID' , 'AxCoefID' , 'JointID'  , 'Dpth'      , 'FillNumM'    , 'MGDpth'    , 'SimplCd'  , 'RNodes'       , 'kp_xr'      , 'mu1'           , 'TwrHtFr'   , 'TwrRe'   , 'RJointID'        , 'IJointID'        , 'COSMID'             , 'CMJointID'        , 'WT_X']
        NUMTAB_FROM_VAL_DIM_VAR = ['NTwInpSt' , 'NumTwrNds' , 'NBlInpSt' , 'DLL_NumTrq' , 'NumBlNds'     , 'NumCases' , 'NHvCoef'  , 'NAxCoef'  , 'NJoints'  , 'NCoefDpth' , 'NFillGroups' , 'NMGDepths' , 1          , 'BldNodes'     , 'kp_total'   , 1               , 'NTwrHt'    , 'NTwrRe'  , 'NReact'          , 'NInterf'         , 'NCOSMs'             , 'NCmass'           , 'NumTurbines']
        NUMTAB_FROM_VAL_VARNAME = ['TowProp'  , 'TowProp'   , 'BldProp'  , 'DLLProp'    , 'BldAeroNodes' , 'Cases'    , 'HvCoefs'  , 'AxCoefs'  , 'Joints'   , 'DpthProp'  , 'FillGroups'  , 'MGProp'    , 'SmplProp' , 'BldAeroNodes' , 'MemberGeom' , 'DampingCoeffs' , 'TowerProp' , 'TowerRe' , 'BaseReactJoints' , 'InterfaceJoints' , 'MemberCosineMatrix' , 'ConcentratedMasses','WindTurbines']
        NUMTAB_FROM_VAL_NHEADER = [2          , 2           , 2          , 2            , 2              , 2          , 2          , 2          , 2          , 2           , 2             , 2           , 2          , 1              , 2            , 2               , 1           , 1         , 2                 , 2                 , 2                    , 2                   ,2]
        NUMTAB_FROM_VAL_TYPE    = ['num'      , 'num'       , 'num'      , 'num'        , 'num'          , 'num'      , 'num'      , 'num'      , 'num'      , 'num'       , 'num'         , 'num'       , 'num'      , 'mix'          , 'num'        , 'num'           , 'num'       , 'num'     , 'mix'             , 'num'             , 'num'                , 'num'               ,'mix']
        NUMTAB_FROM_VAL_DETECT_L = [s.lower() for s in NUMTAB_FROM_VAL_DETECT]

        # NOTE: MJointID1, used by SubDyn and HydroDyn
        NUMTAB_FROM_LAB_DETECT   = ['NumAlf'  , 'F_X'       , 'MemberCd1'    , 'MJointID1' , 'NOutLoc'    , 'NOutCnt'    , 'PropD'             , 'YoungE'            , 'YoungE'          ]
        NUMTAB_FROM_LAB_DIM_VAR  = ['NumAlf'  , 'NKInpSt'   , 'NCoefMembers' , 'NMembers'  , 'NMOutputs'  , 'NMOutputs'  , 'NPropSets'         , 'NPropSets'         , 'NXPropSets'         ]
        NUMTAB_FROM_LAB_VARNAME  = ['AFCoeff' , 'TMDspProp' , 'MemberProp'   , 'Members'   , 'MemberOuts' , 'MemberOuts' , 'MemberSectionProp' , 'MemberSectionProp' , 'MemberSectionProp2' ]
        NUMTAB_FROM_LAB_TYPE     = ['num'     , 'num'       , 'num'          , 'mix'       , 'num'        , 'num'        , 'num'               , 'num'               , 'num'                    ]
        NUMTAB_FROM_LAB_DETECT_L = [s.lower() for s in NUMTAB_FROM_LAB_DETECT]                                         

        FILTAB_FROM_LAB_DETECT   = ['FoilNm' ,'AFNames']
        FILTAB_FROM_LAB_DIM_VAR  = ['NumFoil','NumAFfiles']
        FILTAB_FROM_LAB_VARNAME  = ['FoilNm' ,'AFNames']
        FILTAB_FROM_LAB_DETECT_L = [s.lower() for s in FILTAB_FROM_LAB_DETECT]

        self.data   = []
        self.module = None
        #with open(self.filename, 'r', errors="surrogateescape") as f:
        with open(self.filename, 'r', errors="surrogateescape") as f:
            lines=f.read().splitlines()
        # IF NEEDED> DO THE FOLLOWING FORMATTING:
            #lines = [str(l).encode('utf-8').decode('ascii','ignore') for l in f.read().splitlines()]

        # Fast files start with ! or -
        #if lines[0][0]!='!' and lines[0][0]!='-':
        #    raise Exception('Fast file do not start with ! or -, is it the right format?')

        # Special filetypes
        if self.detectAndReadExtPtfmSE(lines):
            return
        if self.detectAndReadAirfoil(lines):
            return

        # Parsing line by line, storing each line into a dictionary
        i=0    
        nComments  = 0
        nWrongLabels = 0
        allowSpaceSeparatedList=False
        while i<len(lines):
            line = lines[i]
            # OUTLIST Exceptions
            if line.upper().find('ADDITIONAL OUTPUTS')>0 \
            or line.upper().find('MESH-BASED OUTPUTS')>0 \
            or line.upper().find('OUTPUT CHANNELS'   )>0:
                # TODO, lazy implementation so far, MAKE SUB FUNCTION
                parts = re.match(r'^\W*\w+', line)
                if parts:
                    firstword = parts.group(0).strip()
                else:
                    raise NotImplementedError
                remainer  = re.sub(r'^\W*\w+\W*', '', line)
                # Parsing outlist, and then we continue at a new "i" (to read END etc.)
                OutList,i = parseFASTOutList(lines,i+1) 
                d = getDict()
                d['label']   = firstword
                d['descr']   = remainer
                d['tabType'] = TABTYPE_FIL # TODO
                d['value']   = ['']+OutList
                self.data.append(d)
                if i>=len(lines):
                    break

                # --- Here we cheat and force an exit of the input file
                # The reason for this is that some files have a lot of things after the END, which will result in the file being intepreted as a wrong format due to too many comments
                if i+2<len(lines) and lines[i+2].lower().find('bldnd_bladesout')>0:
                    print('>>>Bld Nodal outputs present')
                else:
                    self.data.append(parseFASTInputLine('END of input file (the word "END" must appear in the first 3 columns of this last OutList line)',i+1))
                    self.data.append(parseFASTInputLine('---------------------------------------------------------------------------------------',i+2))
                    break
            elif line.upper().find('SSOUTLIST'   )>0:
                # SUBDYN Outlist doesn not follow regular format
                self.data.append(parseFASTInputLine(line,i))
                # OUTLIST Exception for BeamDyn
                OutList,i = parseFASTOutList(lines,i+1) 
                # TODO
                for o in OutList:
                    d = getDict()
                    d['isComment'] = True
                    d['value']=o
                    self.data.append(d)
                # --- Here we cheat and force an exit of the input file
                self.data.append(parseFASTInputLine('END of input file (the word "END" must appear in the first 3 columns of this last OutList line)',i+1))
                self.data.append(parseFASTInputLine('---------------------------------------------------------------------------------------',i+2))
                break
                
            elif line.upper().find('ADDITIONAL STIFFNESS')>0:
                # TODO, lazy implementation so far, MAKE SUB FUNCTION
                self.data.append(parseFASTInputLine(line,i))
                i +=1
                KDAdd = []
                for _ in range(19):
                    KDAdd.append(lines[i])
                    i +=1
                d = getDict()
                d['label']   = 'KDAdd'   # TODO
                d['tabType'] = TABTYPE_FIL # TODO
                d['value']   = KDAdd
                self.data.append(d)
                if i>=len(lines):
                    break
            elif line.upper().find('DISTRIBUTED PROPERTIES')>0:
                self.data.append(parseFASTInputLine(line,i));
                i+=1;
                self.readBeamDynProps(lines,i)
                return

            # --- Parsing of standard lines: value(s) key comment
            line = lines[i]
            d = parseFASTInputLine(line,i,allowSpaceSeparatedList)

            # --- Handling of special files
            if d['label'].lower()=='kp_total':
                # BeamDyn has weird space speparated list around keypoint definition
                allowSpaceSeparatedList=True
            elif d['label'].lower()=='numcoords':
                # TODO, lazy implementation so far, MAKE SUB FUNCTION
                if isStr(d['value']):
                    if d['value'][0]=='@':
                        # it's a ref to the airfoil coord file
                        pass
                else:
                    if not strIsInt(d['value']): 
                        raise WrongFormatError('Wrong value of NumCoords')
                    if int(d['value'])<=0:
                        pass
                    else:
                        self.data.append(d); i+=1;
                        # 3 comment lines
                        self.data.append(parseFASTInputLine(lines[i],i)); i+=1;
                        self.data.append(parseFASTInputLine(lines[i],i)); i+=1;
                        self.data.append(parseFASTInputLine(lines[i],i)); i+=1;
                        splits=cleanAfterChar(cleanLine(lines[i]),'!').split()
                        # Airfoil ref point
                        try:
                            pos=[float(splits[0]), float(splits[1])]
                        except:
                            raise WrongFormatError('Wrong format while reading coordinates of airfoil reference')
                        i+=1
                        d = getDict()
                        d['label'] = 'AirfoilRefPoint'
                        d['value'] = pos
                        self.data.append(d)
                        # 2 comment lines
                        self.data.append(parseFASTInputLine(lines[i],i)); i+=1;
                        self.data.append(parseFASTInputLine(lines[i],i)); i+=1;
                        # Table of coordinats itself
                        d = getDict()
                        d['label']     = 'AirfoilCoord'
                        d['tabDimVar'] = 'NumCoords'
                        d['tabType']   = TABTYPE_NUM_WITH_HEADERCOM
                        nTabLines = self[d['tabDimVar']]-1  # SOMEHOW ONE DATA POINT LESS
                        d['value'], d['tabColumnNames'],_  = parseFASTNumTable(self.filename,lines[i:i+nTabLines+1],nTabLines,i,1)
                        d['tabUnits'] = ['(-)','(-)']
                        self.data.append(d)
                        break



            #print('label>',d['label'],'<',type(d['label']));
            #print('value>',d['value'],'<',type(d['value']));
            #print(isStr(d['value']))
            #if isStr(d['value']):
            #    print(d['value'].lower() in NUMTAB_FROM_VAL_DETECT_L)

                
            # --- Handling of tables
            if isStr(d['value']) and d['value'].lower() in NUMTAB_FROM_VAL_DETECT_L:
                # Table with numerical values, 
                ii             = NUMTAB_FROM_VAL_DETECT_L.index(d['value'].lower())
                tab_type       = NUMTAB_FROM_VAL_TYPE[ii]
                if tab_type=='num':
                    d['tabType']   = TABTYPE_NUM_WITH_HEADER
                else:
                    d['tabType']   = TABTYPE_MIX_WITH_HEADER
                d['label']     = NUMTAB_FROM_VAL_VARNAME[ii]
                d['tabDimVar'] = NUMTAB_FROM_VAL_DIM_VAR[ii]
                nHeaders       = NUMTAB_FROM_VAL_NHEADER[ii]
                nTabLines=0
                if isinstance(d['tabDimVar'],int):
                    nTabLines = d['tabDimVar']
                else:
                    nTabLines = self[d['tabDimVar']]
                #print('Reading table {} Dimension {} (based on {})'.format(d['label'],nTabLines,d['tabDimVar']));
                d['value'], d['tabColumnNames'], d['tabUnits'] = parseFASTNumTable(self.filename,lines[i:i+nTabLines+nHeaders],nTabLines,i,nHeaders,tableType=tab_type)
                i += nTabLines+nHeaders-1

                # --- Temporary hack for e.g. SubDyn, that has duplicate table, impossible to detect in the current way...
                # So we remove the element form the list one read
                del NUMTAB_FROM_VAL_DETECT[ii]  
                del NUMTAB_FROM_VAL_DIM_VAR[ii] 
                del NUMTAB_FROM_VAL_VARNAME[ii] 
                del NUMTAB_FROM_VAL_NHEADER[ii] 
                del NUMTAB_FROM_VAL_TYPE   [ii] 
                del NUMTAB_FROM_VAL_DETECT_L[ii]  

            elif isStr(d['label']) and d['label'].lower() in NUMTAB_FROM_LAB_DETECT_L:
                ii      = NUMTAB_FROM_LAB_DETECT_L.index(d['label'].lower())
                tab_type       = NUMTAB_FROM_LAB_TYPE[ii]
                # Special case for airfoil data, the table follows NumAlf, so we add d first
                if d['label'].lower()=='numalf':
                    d['tabType']=TABTYPE_NOT_A_TAB
                    self.data.append(d)
                    # Creating a new dictionary for the table
                    d = {'value':None, 'label':'NumAlf', 'isComment':False, 'descr':'', 'tabType':None}
                    i += 1
                d['label']     = NUMTAB_FROM_LAB_VARNAME[ii]
                d['tabDimVar'] = NUMTAB_FROM_LAB_DIM_VAR[ii]
                if d['label'].lower()=='afcoeff' :
                    d['tabType']        = TABTYPE_NUM_WITH_HEADERCOM
                else:
                    if tab_type=='num':
                        d['tabType']   = TABTYPE_NUM_WITH_HEADER
                    else:
                        d['tabType']   = TABTYPE_MIX_WITH_HEADER
                nTabLines = self[d['tabDimVar']]
                #print('Reading table {} Dimension {} (based on {})'.format(d['label'],nTabLines,d['tabDimVar']));
                d['value'], d['tabColumnNames'], d['tabUnits'] = parseFASTNumTable(self.filename,lines[i:i+nTabLines+2],nTabLines,i,2,tableType=tab_type)
                i += nTabLines+1

                # --- Temporary hack for e.g. SubDyn, that has duplicate table, impossible to detect in the current way...
                # So we remove the element form the list one read
                del NUMTAB_FROM_LAB_DETECT[ii]  
                del NUMTAB_FROM_LAB_DIM_VAR[ii] 
                del NUMTAB_FROM_LAB_VARNAME[ii] 
                del NUMTAB_FROM_LAB_TYPE   [ii] 
                del NUMTAB_FROM_LAB_DETECT_L[ii]  

            elif isStr(d['label']) and d['label'].lower() in FILTAB_FROM_LAB_DETECT_L:
                ii             = FILTAB_FROM_LAB_DETECT_L.index(d['label'].lower())
                d['label']     = FILTAB_FROM_LAB_VARNAME[ii]
                d['tabDimVar'] = FILTAB_FROM_LAB_DIM_VAR[ii]
                d['tabType']   = TABTYPE_FIL
                nTabLines = self[d['tabDimVar']]
                #print('Reading table {} Dimension {} (based on {})'.format(d['label'],nTabLines,d['tabDimVar']));
                d['value'] = parseFASTFilTable(lines[i:i+nTabLines],nTabLines,i)
                i += nTabLines-1



            self.data.append(d)
            i += 1
            # --- Safety checks
            if d['isComment']:
                #print(line)
                nComments +=1
            else:
                if hasSpecialChars(d['label']):
                    nWrongLabels +=1
                    #print('label>',d['label'],'<',type(d['label']),line);
                    if i>3: # first few lines may be comments, we allow it
                        #print('Line',i,'Label:',d['label'])
                        raise WrongFormatError('Special Character found in Label: `{}`'.format(d['label']))
                if len(d['label'])==0:
                    nWrongLabels +=1
            if nComments>len(lines)*0.35:
                #print('Comment fail',nComments,len(lines),self.filename)
                raise WrongFormatError('Most lines were read as comments, probably not a FAST Input File')
            if nWrongLabels>len(lines)*0.10:
                #print('Label fail',nWrongLabels,len(lines),self.filename)
                raise WrongFormatError('Too many lines with wrong labels, probably not a FAST Input File')

        # --- PostReading checks
        labels = self.keys()
        duplicates = set([x for x in labels if labels.count(x) > 1])
        if len(duplicates)>0:
            print('[WARN] Duplicate labels found in file: '+self.filename)
            print('       Duplicates: '+', '.join(duplicates))
            print('       It\'s strongly recommended to make them unique! ')
#         except WrongFormatError as e:    
#             raise WrongFormatError('Fast File {}: '.format(self.filename)+'\n'+e.args[0])
#         except Exception as e:    
#             raise e
# #             print(e)
#             raise Exception('Fast File {}: '.format(self.filename)+'\n'+e.args[0])

            
    def toString(self):
        s=''
        # Special file formats, TODO subclass
        if self.module=='ExtPtfm':
            s+='!Comment\n'
            s+='!Comment Flex 5 Format\n'
            s+='!Dimension: {}\n'.format(self['nDOF'])
            s+='!Time increment in simulation: {}\n'.format(self['dt'])
            s+='!Total simulation time in file: {}\n'.format(self['T'])

            s+='\n!Mass Matrix\n'
            s+='!Dimension: {}\n'.format(self['nDOF'])
            s+='\n'.join(''.join('{:16.8e}'.format(x) for x in y) for y in self['MassMatrix'])

            s+='\n\n!Stiffness Matrix\n'
            s+='!Dimension: {}\n'.format(self['nDOF'])
            s+='\n'.join(''.join('{:16.8e}'.format(x) for x in y) for y in self['StiffnessMatrix'])

            s+='\n\n!Damping Matrix\n'
            s+='!Dimension: {}\n'.format(self['nDOF'])
            s+='\n'.join(''.join('{:16.8e}'.format(x) for x in y) for y in self['DampingMatrix'])

            s+='\n\n!Loading and Wave Elevation\n'
            s+='!Dimension: 1 time column -  {} force columns\n'.format(self['nDOF'])
            s+='\n'.join(''.join('{:16.8e}'.format(x) for x in y) for y in self['Loading'])
            return s

        def toStringVLD(val,lab,descr):
            val='{}'.format(val)
            lab='{}'.format(lab)
            if len(val)<13:
                val='{:13s}'.format(val)
            if len(lab)<13:
                lab='{:13s}'.format(lab)
            return val+' '+lab+' - '+descr.strip().strip('-').strip()+'\n'

        for i in range(len(self.data)):
            d=self.data[i]
            if d['isComment']:
                s+='{}'.format(d['value'])
            elif d['tabType']==TABTYPE_NOT_A_TAB:
                if isinstance(d['value'], list):
                    sList=', '.join([str(x) for x in d['value']])
                    s+='{} {} {}'.format(sList,d['label'],d['descr'])
                else:
                    s+=toStringVLD(d['value'],d['label'],d['descr']).strip()
            elif d['tabType']==TABTYPE_NUM_WITH_HEADER:
                s+='{}'.format(' '.join(['{:15s}'.format(s) for s in d['tabColumnNames']]))
                #s+=d['descr'] # Not ready for that
                if d['tabUnits'] is not None:
                    s+='\n'
                    s+='{}'.format(' '.join(['{:15s}'.format(s) for s in d['tabUnits']]))
                if np.size(d['value'],0) > 0 :
                    s+='\n'
                    s+='\n'.join('\t'.join( ('{:15.0f}'.format(x) if int(x)==x else '{:15.8e}'.format(x) )  for x in y) for y in d['value'])
            elif d['tabType']==TABTYPE_MIX_WITH_HEADER:
                s+='{}'.format(' '.join(['{:15s}'.format(s) for s in d['tabColumnNames']]))
                if d['tabUnits'] is not None:
                    s+='\n'
                    s+='{}'.format(' '.join(['{:15s}'.format(s) for s in d['tabUnits']]))
                if np.size(d['value'],0) > 0 :
                    s+='\n'
                    s+='\n'.join('\t'.join('{}'.format(x) for x in y) for y in d['value'])
            elif d['tabType']==TABTYPE_NUM_WITH_HEADERCOM:
                s+='! {}\n'.format(' '.join(['{:15s}'.format(s) for s in d['tabColumnNames']]))
                s+='! {}\n'.format(' '.join(['{:15s}'.format(s) for s in d['tabUnits']]))
                s+='\n'.join('\t'.join('{:15.8e}'.format(x) for x in y) for y in d['value'])
            elif d['tabType']==TABTYPE_FIL:
                #f.write('{} {} {}\n'.format(d['value'][0],d['tabDetect'],d['descr']))
                s+='{} {} {}\n'.format(d['value'][0],d['label'],d['descr']) # TODO?
                s+='\n'.join(fil for fil in d['value'][1:])
            else:
                raise Exception('Unknown table type for variable {}',d)
            if i<len(self.data)-1:
                s+='\n'
        return s

    def _write(self):
        with open(self.filename,'w') as f:
            f.write(self.toString())

    def _toDataFrame(self):
        dfs={}
        # Special types, TODO Subclass
        if self.module=='ExtPtfm':
            nDOF=self['nDOF']
            Cols=['Time_[s]','InpF_Fx_[N]', 'InpF_Fy_[N]', 'InpF_Fz_[N]', 'InpF_Mx_[Nm]', 'InpF_My_[Nm]', 'InpF_Mz_[Nm]']
            Cols+=['CBF_{:03d}_[-]'.format(iDOF+1) for iDOF in np.arange(nDOF)]
            Cols=Cols[:nDOF+1]
            #dfs['Loading']         = pd.DataFrame(data = self['Loading'],columns  = Cols)
            dfs = pd.DataFrame(data = self['Loading'],columns  = Cols)

            #Cols=['SurgeAcc_[m/s]', 'SwayAcc_[m/s]', 'HeaveAcc_[m/s]', 'RollAcc_[rad/s]', 'PitchAcc_[rad/s]', 'YawAcc_[rad/s]']
            #Cols+=['CBQD_{:03d}_[-]'.format(iDOF+1) for iDOF in np.arange(nDOF)]
            #Cols=Cols[:nDOF]
            #dfs['MassMatrix']      = pd.DataFrame(data = self['MassMatrix'], columns=Cols)

            #Cols=['SurgeVel_[m/s]', 'SwayVel_[m/s]', 'HeaveVel_[m/s]', 'RollVel_[rad/s]', 'PitchVel_[rad/s]', 'YawVel_[rad/s]']
            #Cols+=['CBQD_{:03d}_[-]'.format(iDOF+1) for iDOF in np.arange(nDOF)]
            #Cols=Cols[:nDOF]
            #dfs['DampingMatrix']   = pd.DataFrame(data = self['DampingMatrix'], columns=Cols)

            #Cols=['Surge_[m]', 'Sway_[m]', 'Heave_[m]', 'Roll_[rad]', 'Pitch_[rad]', 'Yaw_[rad]']
            #Cols+=['CBQ_{:03d}_[-]'.format(iDOF+1) for iDOF in np.arange(nDOF)]
            #Cols=Cols[:nDOF]
            #dfs['StiffnessMatrix'] = pd.DataFrame(data = self['StiffnessMatrix'], columns=Cols)
            return dfs

        for i in range(len(self.data)): 
            d=self.data[i]
            if d['tabType'] in [TABTYPE_NUM_WITH_HEADER, TABTYPE_NUM_WITH_HEADERCOM, TABTYPE_NUM_NO_HEADER]:
                Val= d['value']
                if d['tabUnits'] is None:
                    Cols=d['tabColumnNames']
                else:
                    Cols=['{}_{}'.format(c,u.replace('(','[').replace(')',']')) for c,u in zip(d['tabColumnNames'],d['tabUnits'])]
                #print(Val)
                #print(Cols)
                if self.getIDSafe('BldFl1Sh(2)')>0:
                    # Hack for blade files, we add the modes
                    x=Val[:,0]
                    Modes=np.zeros((x.shape[0],3))
                    Modes[:,0] = x**2 * self['BldFl1Sh(2)'] \
                               + x**3 * self['BldFl1Sh(3)'] \
                               + x**4 * self['BldFl1Sh(4)'] \
                               + x**5 * self['BldFl1Sh(5)'] \
                               + x**6 * self['BldFl1Sh(6)'] 
                    Modes[:,1] = x**2 * self['BldFl2Sh(2)'] \
                               + x**3 * self['BldFl2Sh(3)'] \
                               + x**4 * self['BldFl2Sh(4)'] \
                               + x**5 * self['BldFl2Sh(5)'] \
                               + x**6 * self['BldFl2Sh(6)'] 
                    Modes[:,2] = x**2 * self['BldEdgSh(2)'] \
                               + x**3 * self['BldEdgSh(3)'] \
                               + x**4 * self['BldEdgSh(4)'] \
                               + x**5 * self['BldEdgSh(5)'] \
                               + x**6 * self['BldEdgSh(6)'] 
                    Val = np.hstack((Val,Modes))
                    Cols = Cols + ['ShapeFlap1_[-]','ShapeFlap2_[-]','ShapeEdge1_[-]']
                  
                elif self.getIDSafe('TwFAM1Sh(2)')>0:
                    # Hack for tower files, we add the modes
                    x=Val[:,0]
                    Modes=np.zeros((x.shape[0],4))
                    Modes[:,0] = x**2 * self['TwFAM1Sh(2)'] \
                               + x**3 * self['TwFAM1Sh(3)'] \
                               + x**4 * self['TwFAM1Sh(4)'] \
                               + x**5 * self['TwFAM1Sh(5)'] \
                               + x**6 * self['TwFAM1Sh(6)'] 
                    Modes[:,1] = x**2 * self['TwFAM2Sh(2)'] \
                               + x**3 * self['TwFAM2Sh(3)'] \
                               + x**4 * self['TwFAM2Sh(4)'] \
                               + x**5 * self['TwFAM2Sh(5)'] \
                               + x**6 * self['TwFAM2Sh(6)'] 
                    Modes[:,2] = x**2 * self['TwSSM1Sh(2)'] \
                               + x**3 * self['TwSSM1Sh(3)'] \
                               + x**4 * self['TwSSM1Sh(4)'] \
                               + x**5 * self['TwSSM1Sh(5)'] \
                               + x**6 * self['TwSSM1Sh(6)'] 
                    Modes[:,3] = x**2 * self['TwSSM2Sh(2)'] \
                               + x**3 * self['TwSSM2Sh(3)'] \
                               + x**4 * self['TwSSM2Sh(4)'] \
                               + x**5 * self['TwSSM2Sh(5)'] \
                               + x**6 * self['TwSSM2Sh(6)'] 
                    Val = np.hstack((Val,Modes))
                    Cols = Cols + ['ShapeForeAft1_[-]','ShapeForeAft2_[-]','ShapeSideSide1_[-]','ShapeSideSide2_[-]']

                name=d['label']
                dfs[name]=pd.DataFrame(data=Val,columns=Cols)
            elif d['tabType'] in [TABTYPE_NUM_BEAMDYN]:
                data = d['value']
                Cols =['Span'] 
                Cols+=['K{}{}'.format(i+1,j+1) for i in range(6) for j in range(6)] 
                Cols+=['M{}{}'.format(i+1,j+1) for i in range(6) for j in range(6)] 
                # Putting the main terms first
                IAll = range(1+36+36)
                IMain= [0] + [i*6+i+1 for i in range(6)] + [i*6+i+37 for i in range(6)]
                IOrg = IMain + [i for i in range(1+36+36) if i not in IMain]
                Cols = [Cols[i] for i in IOrg]
                data = data[:,IOrg]
                name=d['label']
                dfs[name]=pd.DataFrame(data=data,columns=Cols)
        if len(dfs)==1:
            dfs=dfs[list(dfs.keys())[0]]
        return dfs

# --------------------------------------------------------------------------------}
# --- SubReaders /detectors
# --------------------------------------------------------------------------------{
    def detectAndReadExtPtfmSE(self,lines):
        def readmat(n,m,lines,iStart):
            M=np.zeros((n,m))
            for j in np.arange(n):
                i=iStart+j
                M[j,:]=np.array(lines[i].split()).astype(float)
            return M
        if len(lines)<10:
            return False
        if not (lines[0][0]=='!' and lines[1][0]=='!'):
            return False
        if lines[1].lower().find('flex')<0:
            return
        if  lines[2].lower().find('!dimension')<0:
            return
        # --- At this stage we assume it's in the proper format
        self.module='ExtPtfm'
        nDOFCommon = -1
        i=2;
        try:
            while i<len(lines):
                l=lines[i].lower()
                if l.find('!mass')==0:
                    l=lines[i+1]
                    nDOF=int(l.split(':')[1])
                    if nDOF<-1 or nDOF!=nDOFCommon:
                        raise BrokenFormatError('ExtPtfm stiffness matrix nDOF issue. nDOF common: {}, nDOF provided: {}'.format(nDOFCommon,nDOF))
                    self.addKeyVal('MassMatrix',readmat(nDOF,nDOF,lines,i+2))
                    i=i+2+nDOF
                elif l.find('!stiffness')==0:
                    l=lines[i+1]
                    nDOF=int(l.split(':')[1])
                    if nDOF<-1 or nDOF!=nDOFCommon:
                        raise BrokenFormatError('ExtPtfm stiffness matrix nDOF issue nDOF common: {}, nDOF provided: {}'.format(nDOFCommon,nDOF))
                    self.addKeyVal('StiffnessMatrix',readmat(nDOF,nDOF,lines,i+2))
                    i=i+2+nDOF
                elif l.find('!damping')==0:
                    l=lines[i+1]
                    nDOF=int(l.split(':')[1])
                    if nDOF<-1 or nDOF!=nDOFCommon:
                        raise BrokenFormatError('ExtPtfm damping matrix nDOF issue nDOF common: {}, nDOF provided: {}'.format(nDOFCommon,nDOF))
                    self.addKeyVal('DampingMatrix',readmat(nDOF,nDOF,lines,i+2))
                    i=i+2+nDOF
                elif l.find('!loading')==0:
                    try: 
                        nt=int(self['T']/self['dt'])+1
                    except:
                        raise BrokenFormatError('Cannot read loading since time step and simulation time not properly set.')
                    self.addKeyVal('Loading',readmat(nt,nDOFCommon+1,lines,i+2))
                    i=i+nt+2
                elif len(l)>0:
                    if l[0]=='!':
                        if l.find('!dimension')==0:
                            self.addKeyVal('nDOF',int(l.split(':')[1]))
                            nDOFCommon=self['nDOF']
                        elif l.find('!time increment')==0:
                            self.addKeyVal('dt',np.float(l.split(':')[1]))
                        elif l.find('!total simulation time')==0:
                            self.addKeyVal('T',np.float(l.split(':')[1]))
                    else:
                        raise BrokenFormatError('Unexcepted content found on line {}'.format(i))
                i+=1
        except BrokenFormatError as e:
            raise e
        except: 
            raise


        return True
        


    def detectAndReadAirfoil(self,lines):
        if len(lines)<14:
            return False
        # Reading number of tables
        L3 = lines[2].strip().split()
        if len(L3)<=0:
            return False
        if not strIsInt(L3[0]):
            return False
        nTables=int(L3[0])
        # Reading table ID
        L4 = lines[3].strip().split()
        if len(L4)<=nTables:
            return False
        TableID=L4[:nTables]
        if nTables==1:
            TableID=['']
        # Keywords for file format
        KW1=lines[12].strip().split()
        KW2=lines[13].strip().split()
        if len(KW1)>nTables and len(KW2)>nTables:
            if KW1[nTables].lower()=='angle' and KW2[nTables].lower()=='minimum':
                d = getDict(); d['isComment'] = True; d['value'] = lines[0]; self.data.append(d);
                d = getDict(); d['isComment'] = True; d['value'] = lines[1]; self.data.append(d);
                for i in range(2,14):
                    splits = lines[i].split()
                    #print(splits)
                    d = getDict()
                    d['label'] = ' '.join(splits[1:]) # TODO
                    d['descr'] = ' '.join(splits[1:]) # TODO
                    d['value'] = float(splits[0])
                    self.data.append(d)
                #pass
                #for i in range(2,14):
                nTabLines=0
                while 14+nTabLines<len(lines) and  len(lines[14+nTabLines].strip())>0 :
                    nTabLines +=1
                #data = np.array([lines[i].strip().split() for i in range(14,len(lines)) if len(lines[i])>0]).astype(np.float)
                #data = np.array([lines[i].strip().split() for i in takewhile(lambda x: len(lines[i].strip())>0, range(14,len(lines)-1))]).astype(np.float)
                data = np.array([lines[i].strip().split() for i in range(14,nTabLines+14)]).astype(np.float)
                #print(data)
                d = getDict()
                d['label']     = 'Polar'
                d['tabDimVar'] = nTabLines
                d['tabType']   = TABTYPE_NUM_NO_HEADER
                d['value']     = data
                if np.size(data,1)==1+nTables*3:
                    d['tabColumnNames'] = ['Alpha']+[n+l for l in TableID for n in ['Cl','Cd','Cm']]
                    d['tabUnits']       = ['(deg)']+['(-)' , '(-)' , '(-)']*nTables
                elif np.size(data,1)==1+nTables*2:
                    d['tabColumnNames'] = ['Alpha']+[n+l for l in TableID for n in ['Cl','Cd']]
                    d['tabUnits']       = ['(deg)']+['(-)' , '(-)']*nTables
                else:
                    d['tabColumnNames'] = ['col{}'.format(j) for j in range(np.size(data,1))]
                self.data.append(d)
                return True

    def readBeamDynProps(self,lines,iStart):
        nStations=self['station_total']
        M=np.zeros((nStations,1+36+36))
        i=iStart;
        try:
            for j in range(nStations):
                M[j,0]=float(lines[i]); i+=1;
                LL = lines[i:i+6]
                M[j,1:37]=np.array((' '.join(lines[i:i+6])).split()).astype(np.float)
                i+=7
                M[j,37:]=np.array((' '.join(lines[i:i+6])).split()).astype(np.float)
                i+=7
        except: 
            raise WrongFormatError('An error occured while reading section {}/{}'.format(j+1,nStations))
        d = getDict()
        d['label']   = 'BeamProperties'
        d['descr']   = ''
        d['tabType'] = TABTYPE_NUM_BEAMDYN
        d['value']   = M
        self.data.append(d)


# --------------------------------------------------------------------------------}
# --- Helper functions 
# --------------------------------------------------------------------------------{
def isStr(s):
    # Python 2 and 3 compatible
    # Two options below
    # NOTE: all this avoided since we import str from builtins
    # --- Version 2
    #     isString = False;
    #     if(isinstance(s, str)):
    #         isString = True;
    #     try:
    #         if(isinstance(s, basestring)): # todo unicode as well
    #             isString = True;
    #     except NameError:
    #         pass; 
    #     return isString
    # --- Version 1
    #     try: 
    #        basestring # python 2
    #        return isinstance(s, basestring) or isinstance(s,unicode)
    #     except NameError:
    #          basestring=str #python 3
    #     return isinstance(s, str)
   return isinstance(s, str)

def strIsFloat(s):
    #return s.replace('.',',1').isdigit()
    try:
        float(s)
        return True
    except:
        return False

def strIsBool(s):
    return (s.lower() == 'true') or (s.lower() == 'false')

def strIsInt(s):
    s = str(s)
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()    

def hasSpecialChars(s):
    # fast allows for parenthesis
    # For now we allow for - but that's because of BeamDyn geometry members 
    return not re.match("^[\"\'a-zA-Z0-9_()-]*$", s)

def cleanLine(l):
    # makes a string single space separated
    l = l.replace('\t',' ')
    l = ' '.join(l.split())
    l = l.strip()
    return l

def cleanAfterChar(l,c):
    # remove whats after a character
    n = l.find(c);
    if n>0:
        return l[:n]
    else:
        return l

def getDict():
    return {'value':None, 'label':'', 'isComment':False, 'descr':'', 'tabType':TABTYPE_NOT_A_TAB}


def parseFASTInputLine(line_raw,i,allowSpaceSeparatedList=False):
    d = getDict()
    #print(line_raw)
    try:
        # preliminary cleaning (Note: loss of formatting)
        line = cleanLine(line_raw)
        # Comment
        if any(line.startswith(c) for c in ['#','!','--','==']) or len(line)==0:
            d['isComment']=True
            d['value']=line_raw
            return d

        # Detecting lists
        List=[];
        iComma=line.find(',')
        if iComma>0 and iComma<30:
            fakeline=line.replace(' ',',')
            fakeline=re.sub(',+',',',fakeline)
            csplits=fakeline.split(',')
            # Splitting based on comma and looping while it's numbers of booleans
            ii=0
            s=csplits[ii]
            #print(csplits)
            while strIsFloat(s) or strIsBool(s) and ii<len(csplits):
                if strIsInt(s):
                    List.append(int(s))
                elif strIsFloat(s):
                    List.append(float(s))
                elif strIsBool(s):
                    List.append(bool(s))
                else:
                    raise WrongFormatError('Lists of strings not supported.')
                ii =ii+1
                if ii>=len(csplits):
                    raise WrongFormatError('Wrong number of list values')
                s = csplits[ii]
            #print('[INFO] Line {}: Found list: '.format(i),List)
        # Defining value and remaining splits
        if len(List)>=2:
            d['value']=List
            line_remaining=line
            # eating line, removing each values
            for iii in range(ii):
                sValue=csplits[iii]
                ipos=line_remaining.find(sValue)
                line_remaining = line_remaining[ipos+len(sValue):]
            splits=line_remaining.split()
            iNext=0
        else:
            # It's not a list, we just use space as separators
            splits=line.split(' ')
            s=splits[0]

            if strIsInt(s):
                d['value']=int(s)
                if allowSpaceSeparatedList and len(splits)>1:
                    if strIsInt(splits[1]):
                        d['value']=splits[0]+ ' '+splits[1]
            elif strIsFloat(s):
                d['value']=float(s)
            elif strIsBool(s):
                d['value']=bool(s)
            else:
                d['value']=s
            iNext=1
            #import pdb  ; pdb.set_trace();

        # Extracting label (TODO, for now only second split)
        bOK=False
        while (not bOK) and iNext<len(splits):
            # Nasty handling of !XXX: comments
            if splits[iNext][0]=='!' and splits[iNext][-1]==':': 
                iNext=iNext+2
                continue
            # Nasty handling of the fact that sometimes old values are repeated before the label
            if strIsFloat(splits[iNext]):
                iNext=iNext+1
                continue
            else:
                bOK=True
        if bOK:
            d['label']= splits[iNext].strip()
            iNext = iNext+1
        else:
            #print('[WARN] Line {}: No label found -> comment assumed'.format(i+1))
            d['isComment']=True
            d['value']=line_raw
            iNext = len(splits)+1
        
        # Recombining description
        if len(splits)>=iNext+1:
            d['descr']=' '.join(splits[iNext:])
    except WrongFormatError as e:
        raise WrongFormatError('Line {}: '.format(i+1)+e.args[0])
    except Exception as e:
        raise Exception('Line {}: '.format(i+1)+e.args[0])

    return d

def parseFASTOutList(lines,iStart):
    OutList=[]
    i = iStart
    MAX=200
    while i<len(lines) and lines[i].upper().find('END')!=0:
        OutList.append(lines[i]) #TODO better parsing
        #print('OutList',lines[i])
        i += 1
        if i-iStart>MAX :
            raise Exception('More that 200 lines found in outlist')
        if i>=len(lines):
            print('[WARN] End of file reached while reading Outlist')
    #i=min(i+1,len(lines))
    return OutList,iStart+len(OutList)


def extractWithinParenthesis(s):
    mo = re.search(r'\((.*)\)', s)
    if mo:
        return mo.group(1)
    return ''

def extractWithinBrackets(s):
    mo = re.search(r'\((.*)\)', s)
    if mo:
        return mo.group(1)
    return ''

def detectUnits(s,nRef):
    nPOpen=s.count('(')
    nPClos=s.count(')')
    nBOpen=s.count('[')
    nBClos=s.count(']')

    sep='!#@#!'
    if (nPOpen == nPClos) and (nPOpen>=nRef):
        #that's pretty good
        Units=s.replace('(','').replace(')',sep).split(sep)[:-1]
    elif (nBOpen == nBClos) and (nBOpen>=nRef):
        Units=s.replace('[','').replace(']',sep).split(sep)[:-1]
    else:
        Units=s.split()
    return Units


def parseFASTNumTable(filename,lines,n,iStart,nHeaders=2,tableType='num'):
    Tab = None
    ColNames = None
    Units = None

    if len(lines)!=n+nHeaders:
        raise BrokenFormatError('Not enough lines in table: {} lines instead of {}\nFile:{}'.format(len(lines)-nHeaders,n,filename))

    if nHeaders<1:
        raise NotImplementedError('Reading FAST tables with no headers not implemented yet')

    try:
        if nHeaders>=1:
            # Extract column names
            i = 0
            sTmp = cleanLine(lines[i])
            sTmp = cleanAfterChar(sTmp,'[')
            if sTmp.startswith('!'):
                sTmp=sTmp[1:].strip()
            ColNames=sTmp.split()
        if nHeaders>=2:
            # Extract units
            i = 1
            sTmp = cleanLine(lines[i])
            if sTmp.startswith('!'):
                sTmp=sTmp[1:].strip()

            Units = detectUnits(sTmp,len(ColNames))
            Units = ['({})'.format(u.strip()) for u in Units]
            # Forcing user to match number of units and column names
            if len(ColNames) != len(Units):
                print(ColNames)
                print(Units)
                raise Exception('Number of column names different from number of units in table')

        nCols=len(ColNames)

        if tableType=='num':
            Tab = np.zeros((n, nCols))
            for i in range(nHeaders,n+nHeaders):
                l = lines[i].lower()
                v = l.split()
                if len(v) > nCols:
                    print('[WARN] {}: Line {}: number of data different from number of column names'.format(filename, iStart+i+1))
                if len(v) < nCols:
                    raise Exception('Number of data is lower than number of column names')
                # Accounting for TRUE FALSE and converting to float
                v = [s.replace('true','1').replace('false','0').replace('noprint','0').replace('print','1') for s in v]
                v = [float(s) for s in v[0:nCols]]
                if len(v) < nCols:
                    raise Exception('Number of data is lower than number of column names')
                Tab[i-nHeaders,:] = v
        elif tableType=='mix':
            # a mix table contains a mixed of strings and floats
            # For now, we are being a bit more relaxed about the number of columns
            Tab = np.zeros((n, nCols)).astype(object)
            for i in range(nHeaders,n+nHeaders):
                l = lines[i]
                v = l.split()
                if len(v) != nCols:
                    print('[WARN] {}: Line {}: Number of data is different than number of column names'.format(filename,iStart+1+i))
                v=v[0:min(len(v),nCols)]
                Tab[i-nHeaders,0:len(v)] = v
        else:
            raise Exception('Unknown table type')
            
    except Exception as e:    
        raise BrokenFormatError('Line {}: {}'.format(iStart+i+1,e.args[0]))
    return Tab, ColNames, Units


def parseFASTFilTable(lines,n,iStart):
    Tab = []
    try:
        i=0
        if len(lines)!=n:
            raise WrongFormatError('Not enough lines in table: {} lines instead of {}'.format(len(lines),n))
        for i in range(n):
            l = lines[i].split()
            #print(l[0].strip())
            Tab.append(l[0].strip())
            
    except Exception as e:    
        raise Exception('Line {}: '.format(iStart+i+1)+e.args[0])
    return Tab


if __name__ == "__main__":
    B=FASTInFile('BeamDyn_Blade.dat')






import os
# for encoding detection:
#import codecs
#import chardet 

class WrongFormatError(Exception):
    pass

class EmptyFileError(Exception):
    pass

class BrokenFormatError(Exception):
    pass

class BrokenReaderError(Exception):
    pass

try: #Python3
    FileNotFoundError=FileNotFoundError
except NameError: # Python2
    FileNotFoundError = IOError

class File(dict):
    def __init__(self,filename=None,**kwargs):
        self._size=None
        self._encoding=None
        if filename:
            ### If there is a new filename, replace the object variable
            self.filename = filename
            ### If the filename is provided, read the file
            self.read(**kwargs)
        else:
            self.filename = None

    def read(self, filename=None, **kwargs):
        if filename:
            self.filename = filename
        if not self.filename:
            raise Exception('No filename provided')
        if not os.path.isfile(self.filename):
            raise OSError(2,'File not found:',self.filename)
        if os.stat(self.filename).st_size == 0:
            raise EmptyFileError('File is empty:',self.filename)
        # Calling children function
        self._read(**kwargs)

    def write(self, filename=None):
        if filename:
            self.filename = filename
        if not self.filename:
            raise Exception('No filename provided')
        # Calling children function
        self._write()

    def toDataFrame(self):
        return self._toDataFrame()

    # --------------------------------------------------------------------------------
    # --- Properties
    # --------------------------------------------------------------------------------
    @property
    def size(self):
        if self._size is None:
            self._size = os.path.getsize(self.filename)
        return self._size

    #@property
    #def encoding(self):
    #    """  Detects encoding"""
    #    if self._encoding is None:
    #        byts = min(32, self.size)
    #        with open(self.filename, 'rb') as f:
    #            raw = f.read(byts)
    #        if raw.startswith(codecs.BOM_UTF8):
    #            self._encoding = 'utf-8-sig'
    #        else:
    #            result = chardet.detect(raw)
    #            self._encoding = result['encoding']
    #    return self._encoding


    # --------------------------------------------------------------------------------}
    # --- Helper methods
    # --------------------------------------------------------------------------------{
    
    # --------------------------------------------------------------------------------
    # --- Sub class methods 
    # --------------------------------------------------------------------------------
    def _read(self,**kwargs):
        raise NotImplementedError("Method must be implemented in the subclass")

    def _write(self):
        raise NotImplementedError("Method must be implemented in the subclass")

    def _toDataFrame(self):
        raise NotImplementedError("Method must be implemented in the subclass")

    def _fromDataFrame(self):
        raise NotImplementedError("Method must be implemented in the subclass")

    def _fromDictionary(self):
        raise NotImplementedError("Method must be implemented in the subclass")

    def _fromFile(self):
        raise NotImplementedError("Method must be implemented in the subclass")

    # --------------------------------------------------------------------------------
    # --- Static methods
    # --------------------------------------------------------------------------------
    @staticmethod
    def defaultExtension():
        raise NotImplementedError("Method must be implemented in the subclass")

    @staticmethod
    def formatName():
        raise NotImplementedError("Method must be implemented in the subclass")

    @classmethod
    def isRightFormat(cls,filename):
        """ Tries to open a file, return true and the file if it succeeds """
        #raise NotImplementedError("Method must be implemented in the subclass")
        try:
            F=cls(filename=filename)
            return True,F
        except MemoryError:
            raise
        except WrongFormatError:
            return False,None
        except:
            raise

    def test_write_read(self,bDelete=False):
        """ Test that we can write and then read what we wrote
        NOTE: this does not check that what we read is the same..
        """
        # --- First, test write function (assuming read)
        try:
            f,ext=os.path.splitext(self.filename)
            filename_out = f+'_TMP'+ext
            self.write(filename_out)
        except Exception as e:
            raise Exception('Error writing what we read\n'+e.args[0])
        # --- Second,  re-read what we wrote
        try:
            self.read(filename_out)
        except Exception as e:
            raise Exception('Error reading what we wrote\n'+e.args[0])
        if bDelete:
            os.remove(filename_out)
        return filename_out

    def test_ascii(self,bCompareWritesOnly=False,bDelete=True):
        # compare ourselves (assuming read has occured) with what we write

        f,ext=os.path.splitext(self.filename)
        # --- Perform a simple write/read test
        filename_out=self.test_write_read()

        # --- Perform ascii comparison (and delete if success)
        if bCompareWritesOnly:
            f1 = filename_out
            f2 = f+'_TMP2'+ext
            self.write(f2)
        else:
            f1 = self.filename
            f2 = filename_out
        bStat=ascii_comp(f1,f2,bDelete=bDelete)

        if bStat:
            if bCompareWritesOnly and bDelete:
                os.remove(f1)
        else:
            raise Exception('The ascii content of {} and {} are different'.format(f1,f2))

# --------------------------------------------------------------------------------}
# --- Helper functions
# --------------------------------------------------------------------------------{
def isBinary(filename):
    with open(filename, 'r') as f:
        try:
            f.readline()
            return False
        except UnicodeDecodeError:
            return True

def ascii_comp(file1,file2,bDelete=False):
    """ Compares two ascii files line by line.
    Comparison is done ignoring multiple white spaces for now"""
    # --- Read original as ascii
    with open(file1, 'r') as f1:
        lines1 = f1.read().splitlines();
        lines1 = '|'.join([l.replace('\t',' ').strip() for l in lines1])
        lines1 = ' '.join(lines1.split())
    # --- Read second file as ascii
    with open(file2, 'r') as f2:
        lines2 = f2.read().splitlines();
        lines2 = '|'.join([l.replace('\t',' ').strip() for l in lines2])
        lines2 = ' '.join(lines2.split())

    if lines1 == lines2:
        if bDelete:
            os.remove(file2)
        return True
    else:
        return False



from numpy import zeros, array
import numpy as np

# This variable tree contains all parameters required to create a FAST model
# for FAST versions 7 and 8.


##########################
### Main input file ######
##########################
# Runtime Options
class runtime_options(object):
    def __init__(self):
         self.echo = False
         self.RandSeed1 = np.random.uniform(1, 1e8)
         self.RandSeed2 = '"RANLUX"'
         self.WrBHHTP = False
         self.WrFHHTP = False
         self.WrADHH = False
         self.WrADFF = True
         self.WrBLFF = False
         self.WrADTWR = False
         self.WrFMTFF = False
         self.WrACT = False
         self.Clockwise = True
         self.ScaleIEC = 0

# Turbine/Model Specifications
class tmspecs(object):
    def __init__(self):
         self.NumGrid_Z = 30
         self.NumGrid_Y = 30
         self.TimeStep = 0.05
         self.AnalysisTime = 630
         self.UsableTime = '"ALL"'
         self.HubHt = 90
         self.GridHeight = 138
         self.GridWidth = 138
         self.VFlowAng = 0
         self.HFlowAng = 0

# Meteorological Boundary Conditions
class metboundconds(object):
    def __init__(self):
         self.TurbModel = '"USRINP"'
         self.UserFile = 'turbulence_user.inp'
         self.IECstandard = '"1-ED3"'
         self.IECturbc = '"B"'
         self.IEC_WindType = '"NTM"'
         self.ETMc = '"default"'
         self.WindProfileType = '"USR"'
         self.ProfileFile = 'shear.profile'
         self.RefHt = 90
         self.URef = 12
         self.ZJetMax = '"default"'
         self.PLExp = '"default"'
         self.Z0 = '"default"'

# Non-IEC Meteorological Boundary Conditions
class noniecboundconds(object):
    def __init__(self):
         self.Latitude = '"default"'
         self.RICH_NO = 0
         self.UStar = '"default"'
         self.ZI = '"default"'
         self.PC_UW = -0.85
         self.PC_UV = 0.15
         self.PC_VW = -0.1

# Spatial Coherence Parameters
class spatialcoherance(object):
    def __init__(self):
         self.SCMod1 = '"GENERAL"'
         self.SCMod2 = '"GENERAL"'
         self.SCMod3 = '"GENERAL"'
         self.InCDec1 = [13.75, 0.04]
         self.InCDec2 = [9.85, 0.0015]
         self.InCDec3 = [9.5, 0.003]
         self.CohExp = 0.5

# Coherent Turbulence Scaling Parameters
class coherentTurbulence(object):
    def __init__(self):
         self.CTEventPath = '"Y:\Wind\Archive\Public\Projects\KH_Billow\EventData"'
         self.CTEventFile = '"Random"'
         self.Randomize = True
         self.DistScl = 1
         self.CTLy = 0.5
         self.CTLz = 0.5
         self.CTStartTime = 30

class turbsiminputs(object):
    def __init__(self):
         self.runtime_options = runtime_options()
         self.tmspecs = tmspecs()
         self.metboundconds = metboundconds()
         self.noniecboundconds = noniecboundconds()
         self.spatialcoherance = spatialcoherance()
         self.coherentTurbulence = coherentTurbulence()

##########################
### Turbulance input file ######
##########################




from __future__ import print_function
from ROSCO_toolbox.ofTools.fast_io.output_processing import output_processing
import ROSCO_toolbox

def FAST_IO_timeseries(fname):
    # interface to FAST_IO data load
    try:
        test = ROSCO_toolbox.__file__
    except:
        print('WARNING: ROSCO_toolbox required for wisdem.aeroelasticse.FAST_post.FAST_IO_timeseries')
    
    fast_out = output_processing.output_processing()
    fast_data = fast_out.load_fast_out(fname, verbose=True)[0]
    return fast_data



''' 
A number of the file processing tools used here were provided by or modified from: https://github.com/ebranlard/weio. 

Functions:
--------
run_openfast
load_fast_out
load_ascii_output
load_binary_output
trim_output
'''
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import transforms
from itertools import takewhile
import struct

from ROSCO_toolbox.ofTools.util import spectral

class output_processing():
    '''
    Some plotting utilities for OpenFAST data. 

    Methods:
    load_fast_out
    plot_fast_out
    plot_spectral
    '''

    def __init__(self, filenames=[], cases=None, tmin=None, tmax=None, verbose=False):
        
        # Define generic plotting cases
        if cases:
            self.plot_cases=cases
        else:
            self.plot_cases = {'Baseline': ['Wind1VelX', 'GenPwr', 'RotSpeed', 'BldPitch1', 'GenTq']}

        if len(filenames) > 0:
            self.load_fast_out(filenames, tmin=tmin, tmax=tmax, verbose=verbose)

    def load_fast_out(self, filenames, tmin=None, tmax=None, verbose=False):
        """Load a FAST binary or ascii output file
        
        Parameters
        ----------
        filenames : list
            list of filenames
        tmin : float, optional
            initial time to trim output data to 
        tmax : float, optional
            final data to trim output data to
        verbose : bool, optional
            Print updates
        
        Returns
        -------
        fastout: list
            List of dictionaries containing OpenFAST output data.
        """
        if type(filenames) is str:
            filenames = [filenames]
            
        # data = []
        # info = []
        try:
            self.fastout
        except AttributeError:
            self.fastout = []
        for i, filename in enumerate(filenames):
            assert os.path.isfile(filename), "File, %s, does not exists" % filename
            with open(filename, 'r') as f:
                if verbose:
                    print('Loading data from {}'.format(filename))
                try:
                    f.readline()
                except UnicodeDecodeError:
                    data, info = load_binary_output(filename)
                else:
                    data, info = load_ascii_output(filename)

            # Build dictionary
            fast_data = dict(zip(info['channels'],data.T))
            fast_data['meta'] = info
            fast_data['meta']['filename'] = filename
            self.fastout.append(fast_data)

        # Trim outputs
        if (tmin) or (tmax):
            trim_output(self.fastout, tmin=tmin, tmax=tmax, verbose=verbose)

        # return fastout
        return self.fastout

    def plot_fast_out(self, fastout=None, cases=None, showplot=True, fignum=None, xlim=None):
        '''
        Plots OpenFAST outputs for desired channels

        Parameters:
        -----------
        cases : dict
            Dictionary of lists containing desired outputs
        fastout : list
            List of dictionaries of OpenFAST output information, output from load_fast_out
        showplot: bool, optional
            Show the plot
        fignum: int, optional
            Define figure number. Note: Should only be used when plotting a singular case. 

        Returns:
        --------
        figlist: list
            list of figure handles
        axeslist: list
            list of axes handles
        '''

        if not fastout:
            try:
                fastout = self.fastout
            except:
                raise AttributeError('Cannot plot OpenFAST output data before it is loaded with load_fast_out.')
        if not cases:
            cases = self.plot_cases

        figlist = []
        axeslist = []
        # Plot cases
        for case in cases.keys():
            # channels to plot
            channels = cases[case]
            # instantiate plot and legend
            fig, axes = plt.subplots(len(channels), 1, sharex=True, num=fignum, constrained_layout=True)
            myleg = []
            for fast_out in fastout:      # Multiple channels
                # write legend
                Time = fast_out['Time']
                myleg.append(fast_out['meta']['name'])
                if len(channels) > 1:  # Multiple channels
                    for axj, channel_tup in zip(axes, channels):
                        if not isinstance(channel_tup, tuple):
                            channel_tup = (channel_tup,)
                        for cidx, channel in enumerate(channel_tup):
                            try:
                                # plot
                                axj.plot(Time, fast_out[channel])
                                # label
                                unit_idx = fast_out['meta']['channels'].index(channel)
                            except:
                                print('{} is not available as an output channel.'.format(channel))
                        axj.set_ylabel('{:^} \n ({:^})'.format(
                            '\n'.join(channel_tup),
                            fast_out['meta']['attribute_units'][unit_idx]))
                        axj.grid(True)
                    axes[0].set_title(case)
                    
                else:                       # Single channel
                    if not isinstance(channels[0], tuple):
                        channel_tup = (channels[0],)
                    else:
                        channel_tup = channels[0]
                    for cidx, channel in enumerate(channel_tup):
                        try:
                            # plot
                            axes.plot(Time, fast_out[channel])
                            # label
                            channel.replace(' ','\n')
                            axes.grid(True)
                            axes.set_title(case)
                        except:
                            print('{} is not available as an output channel.'.format(channel))
                        axes.set_ylabel('{:^} \n ({:^})'.format(
                            '\n'.join(channel_tup),
                            fast_out['meta']['attribute_units'][unit_idx]))
                
                plt.legend(myleg, loc='upper center', bbox_to_anchor=(
                    0.5, 0.0), borderaxespad=2, ncol=len(fastout))

            
            fig.align_ylabels()
            figlist.append(fig)
            axeslist.append(axes)

            if xlim:
                plt.xlim(xlim)

        if showplot:
            plt.show()

        return figlist, axeslist

    def plot_spectral(self, fastout=None, cases=None,
                      averaging='None', averaging_window='Hann', detrend=False, nExp=None,
                      show_RtSpeed=False, RtSpeed_idx=None,
                      add_freqs=None, add_freq_labels=None,
                      showplot=False, fignum=None):
        '''
        Plots OpenFAST outputs for desired channels

        Parameters:
        -----------
        fastout : dict
            Dictionary of OpenFAST output information, output from load_fast_out
        cases : list of tuples (str, int)
            Dictionary of lists containing desired outputs. 
            Of the format (channel, case), i.e. [('RotSpeed', 0)]
        averaging: str, optional
            PSD averaging method. None, Welch
        averaging_window: str, optional
            PSD averaging window method. Hamming, Hann, Rectangular
        detrend: bool, optional
            Detrend data?
        nExp: float, optional
            Exponent for hamming windowing
        show_RtSpeed: Bool, optional
            Plot 1p and 3p rotor speeds for simulation cases plotted
        RtSpeed_idx: ind, optional
            Specify the index for the simulation case that the rotor speed is plotted from. 
        add_freqs: list, optional
            List of floats containing additional frequencies to plot lines of
        add_freq_labels: list, optional
            List of strings to label add_freqs
        showplot: bool, optional
            Show the plot
        fignum: int, optional
            Define figure number. Note: Should only be used when plotting a singular case. 

        Returns:
        -------
        fig, ax - corresponds to generated figure
        '''

        if not fastout:
            try:
                fastout = self.fastout
            except:
                raise AttributeError('Cannot plot OpenFAST output data before it is loaded with load_fast_out.')
        if not cases:
            cases=self.plot_cases

        if averaging.lower() not in ['none', 'welch']:
            raise ValueError('{} is not a supported averaging method.'.format(averaging))

        if averaging_window.lower() not in ['hamming', 'hann', 'rectangular']:
            raise ValueError('{} is not a supported averaging window.'.format(averaging_window))

        fig, ax = plt.subplots(num=fignum)

        leg = []
        for channel, run in cases:
            try:
                # Find time
                Time = fastout[run]['Time']
                # Load PSD
                fq, y, info = spectral.fft_wrap(
                    Time, fastout[run][channel], averaging=averaging, averaging_window=averaging_window, detrend=detrend, nExp=nExp)
                # Plot data
                plt.loglog(fq, y, label='{}, run: {}'.format(channel, str(run)))
            except:
                print('{} is not an available channel in run {}'.format(channel, str(run)))
        # Show rotor speed range (1P, 3P, 6P?)
        if show_RtSpeed:
            if not RtSpeed_idx:
                RtSpeed_idx = [0]
                print('No rotor speed run indices defined, plotting spectral range for the first run only.')
            for rt_idx in RtSpeed_idx:
                f_1P_min = min(fastout[rt_idx]['RotSpeed']) / 60.  # Hz
                f_1P_max = max(fastout[rt_idx]['RotSpeed']) / 60.
                f_3P_min = f_1P_min*3
                f_3P_max = f_1P_max*3
                f_6P_min = f_1P_min*6
                f_6P_max = f_1P_max*6
                plt.axvspan(f_1P_min, f_1P_max, alpha=0.5, color=[0.7, 0.7, 0.7], label='1P')
                plt.axvspan(f_3P_min, f_3P_max, alpha=0.5, color=[0.8, 0.8, 0.8], label='3P')
                # if f_6P_min<1.:
                #     plt.axvspan(f_6P_min, f_6P_max, alpha=0.5, color=[0.9,0.9,0.9], label='6P')

        # Add specific frequencies if desired
        if add_freqs:
            co = np.linspace(0.3, 0.0, len(add_freqs))
            if add_freq_labels is None:
                add_freq_labels = [None]*len(add_freqs)
            for freq, flabel, c in zip(add_freqs, add_freq_labels, co):
                plt.axvline(freq, color=[c, c, c])  # , label=flabel)
                trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
                plt.text(freq+(10**np.floor(np.log10(freq))/10), 0.01, flabel, transform=trans)

        # Formatting
        plt.legend(loc='best')
        if len(list(cases)) > 1:
            plt.ylabel('PSD')
        else:
            unit_idx = fastout[run]['meta']['channels'].index(channel)
            plt.ylabel(
                'PSD ({}$^2$/Hz)'.format(fastout[run]['meta']['attribute_units'][unit_idx]))
        plt.xlabel('Frequency (Hz)')
        plt.grid(True)

        if showplot:
            plt.show()

        return fig, ax


def load_ascii_output(filename):
    '''
    Load FAST ascii output file 
    
    Parameters
    ----------
    filename : str
        filename
    
    Returns
    -------
    data : ndarray
        data values
    info : dict
        info containing:
            - name: filename
            - description: description of dataset
            - channels: list of attribute names
            - attribute_units: list of attribute units
    '''
    with open(filename) as f:
        info = {}
        info['name'] = os.path.splitext(os.path.basename(filename))[0]
        # Header is whatever is before the keyword `time`
        in_header = True
        header = []
        while in_header:
            l = f.readline()
            if not l:
                raise Exception('Error finding the end of FAST out file header. Keyword Time missing.')
            in_header= (l+' dummy').lower().split()[0] != 'time'
            if in_header:
                header.append(l)
            else:
                info['description'] = header
                info['channels'] = l.split()
                info['attribute_units'] = [unit[1:-1] for unit in f.readline().split()]

        # Data, up to end of file or empty line (potential comment line at the end)
        data = np.array([l.strip().split() for l in takewhile(lambda x: len(x.strip())>0, f.readlines())]).astype(np.float)
        return data, info


def load_binary_output(filename, use_buffer=True):
    """
            
    Info about ReadFASTbinary.m:
    
    Original Author: Bonnie Jonkman, National Renewable Energy Laboratory
    (c) 2012, National Renewable Energy Laboratory
    Edited for FAST v7.02.00b-bjj  22-Oct-2012
    03/09/15: Ported from ReadFASTbinary.m by Mads M Pedersen, DTU Wind
    10/24/18: Low memory/buffered version by E. Branlard, NREL
    18/01/19: New file format for exctended channels, by E. Branlard, NREL
    11/4/19: Implemented in ROSCO toolbox by N. Abbas, NREL
    8/6/20: Synced between rosco toolbox and weio by P Bortolotti, NREL

    Parameters
    ----------
    filename : str
        filename
    Returns
    -------
    data : ndarray
        data values
    info : dict
        info containing:
            - name: filename
            - description: description of dataset
            - channels: list of attribute names
            - attribute_units: list of attribute units
    """
    def fread(fid, n, type):
        fmt, nbytes = {'uint8': ('B', 1), 'int16':('h', 2), 'int32':('i', 4), 'float32':('f', 4), 'float64':('d', 8)}[type]
        return struct.unpack(fmt * n, fid.read(nbytes * n))

    def freadRowOrderTableBuffered(fid, n, type_in, nCols, nOff=0, type_out='float64'):
        """ 
        Reads of row-ordered table from a binary file.
        Read `n` data of type `type_in`, assumed to be a row ordered table of `nCols` columns.
        Memory usage is optimized by allocating the data only once.
        Buffered reading is done for improved performances (in particular for 32bit python)
        `nOff` allows for additional column space at the begining of the storage table.
        Typically, `nOff=1`, provides a column at the beginning to store the time vector.
        @author E.Branlard, NREL
        """
        fmt, nbytes = {'uint8': ('B', 1), 'int16':('h', 2), 'int32':('i', 4), 'float32':('f', 4), 'float64':('d', 8)}[type_in]
        nLines          = int(n/nCols)
        GoodBufferSize  = 4096*40
        nLinesPerBuffer = int(GoodBufferSize/nCols)
        BufferSize      = nCols * nLinesPerBuffer
        nBuffer         = int(n/BufferSize)
        # Allocation of data
        data = np.zeros((nLines,nCols+nOff), dtype = type_out)
        # Reading
        try:
            nIntRead   = 0
            nLinesRead = 0
            while nIntRead<n:
                nIntToRead = min(n-nIntRead, BufferSize)
                nLinesToRead = int(nIntToRead/nCols)
                Buffer = np.array(struct.unpack(fmt * nIntToRead, fid.read(nbytes * nIntToRead)))
                Buffer = Buffer.reshape(-1,nCols)
                data[ nLinesRead:(nLinesRead+nLinesToRead),  nOff:(nOff+nCols)  ] = Buffer
                nLinesRead = nLinesRead + nLinesToRead
                nIntRead   = nIntRead   + nIntToRead
        except:
            raise Exception('Read only %d of %d values in file:' % (nIntRead, n, filename))
        return data


    FileFmtID_WithTime              = 1 # File identifiers used in FAST
    FileFmtID_WithoutTime           = 2
    FileFmtID_NoCompressWithoutTime = 3
    FileFmtID_ChanLen_In            = 4

    with open(filename, 'rb') as fid:
        #----------------------------        
        # get the header information
        #----------------------------

        FileID = fread(fid, 1, 'int16')[0]  #;             % FAST output file format, INT(2)

        if FileID not in [FileFmtID_WithTime, FileFmtID_WithoutTime, FileFmtID_NoCompressWithoutTime, FileFmtID_ChanLen_In]:
            raise Exception('FileID not supported {}. Is it a FAST binary file?'.format(FileID))

        if FileID == FileFmtID_ChanLen_In: 
            LenName = fread(fid, 1, 'int16')[0] # Number of characters in channel names and units
        else:
            LenName = 10                    # Default number of characters per channel name

        NumOutChans = fread(fid, 1, 'int32')[0]  #;             % The number of output channels, INT(4)
        NT = fread(fid, 1, 'int32')[0]  #;             % The number of time steps, INT(4)

        if FileID == FileFmtID_WithTime:
            TimeScl = fread(fid, 1, 'float64')  #;           % The time slopes for scaling, REAL(8)
            TimeOff = fread(fid, 1, 'float64')  #;           % The time offsets for scaling, REAL(8)
        else:
            TimeOut1 = fread(fid, 1, 'float64')  #;           % The first time in the time series, REAL(8)
            TimeIncr = fread(fid, 1, 'float64')  #;           % The time increment, REAL(8)

        if FileID == FileFmtID_NoCompressWithoutTime:
            ColScl = np.ones ((NumOutChans, 1)) # The channel slopes for scaling, REAL(4)
            ColOff = np.zeros((NumOutChans, 1)) # The channel offsets for scaling, REAL(4)
        else:
            ColScl = fread(fid, NumOutChans, 'float32')  # The channel slopes for scaling, REAL(4)
            ColOff = fread(fid, NumOutChans, 'float32')  # The channel offsets for scaling, REAL(4)

        LenDesc      = fread(fid, 1, 'int32')[0]  #;  % The number of characters in the description string, INT(4)
        DescStrASCII = fread(fid, LenDesc, 'uint8')  #;  % DescStr converted to ASCII
        DescStr      = "".join(map(chr, DescStrASCII)).strip()

        ChanName = []  # initialize the ChanName cell array
        for iChan in range(NumOutChans + 1):
            ChanNameASCII = fread(fid, LenName, 'uint8')  #; % ChanName converted to numeric ASCII
            ChanName.append("".join(map(chr, ChanNameASCII)).strip())

        ChanUnit = []  # initialize the ChanUnit cell array
        for iChan in range(NumOutChans + 1):
            ChanUnitASCII = fread(fid, LenName, 'uint8')  #; % ChanUnit converted to numeric ASCII
            ChanUnit.append("".join(map(chr, ChanUnitASCII)).strip()[1:-1])

        # -------------------------
        #  get the channel time series
        # -------------------------

        nPts = NT * NumOutChans  #;           % number of data points in the file

        if FileID == FileFmtID_WithTime:
            PackedTime = fread(fid, NT, 'int32')  #; % read the time data
            cnt = len(PackedTime)
            if cnt < NT:
                raise Exception('Could not read entire %s file: read %d of %d time values' % (filename, cnt, NT))

        if use_buffer:
            # Reading data using buffers, and allowing an offset for time column (nOff=1)
            if FileID == FileFmtID_NoCompressWithoutTime:
                data = freadRowOrderTableBuffered(fid, nPts, 'float64', NumOutChans, nOff=1, type_out='float64')
            else:
                data = freadRowOrderTableBuffered(fid, nPts, 'int16', NumOutChans, nOff=1, type_out='float64')
        else:
            # NOTE: unpacking huge data not possible on 32bit machines
            if FileID == FileFmtID_NoCompressWithoutTime:
                PackedData = fread(fid, nPts, 'float64')  #; % read the channel data
            else:
                PackedData = fread(fid, nPts, 'int16')  #; % read the channel data

            cnt = len(PackedData)
            if cnt < nPts:
                raise Exception('Could not read entire %s file: read %d of %d values' % (filename, cnt, nPts))
            data = np.array(PackedData).reshape(NT, NumOutChans)
            del PackedData

    if FileID == FileFmtID_WithTime:
        time = (np.array(PackedTime) - TimeOff) / TimeScl;
    else:
        time = TimeOut1 + TimeIncr * np.arange(NT)

    # -------------------------
    #  Scale the packed binary to real data
    # -------------------------
    if use_buffer:
        # Scaling Data
        for iCol in range(NumOutChans):
            if np.isnan(ColScl[iCol]) and np.isnan(ColOff[iCol]):
                data[:,iCol+1] = 0 # probably due to a division by zero in Fortran
            else:
                data[:,iCol+1] = (data[:,iCol+1] - ColOff[iCol]) / ColScl[iCol]
        # Adding time column
        data[:,0] = time
    else:
        # NOTE: memory expensive due to time conversion, and concatenation
        data = (data - ColOff) / ColScl
        data = np.concatenate([time.reshape(NT, 1), data], 1)

    info = {'name': os.path.splitext(os.path.basename(filename))[0],
            'description': DescStr,
            'channels': ChanName,
            'attribute_units': ChanUnit}
    return data, info

def trim_output(fast_data, tmin=None, tmax=None, verbose=False):
    '''
    Trim loaded fast output data 
    Parameters
    ----------
    fast_data : list
        List of all output data from load_fast_out (list containing dictionaries)
    tmin : float, optional
        start time
    tmax : float, optional
        end time
    
    Returns
    -------
    fast_data : list
        list of dictionaries containing trimmed fast output data
    '''
    if isinstance(fast_data, dict):
        fast_data = [fast_data]



    # initial time array and associated index
    for fd in fast_data:
        if verbose:
            if tmin: 
                tmin_v = str(tmin) + ' seconds'
            else:  
                tmin_v = 'the beginning'
            if tmax: 
                tmax_v = str(tmax) + ' seconds'
            else: 
                tmax_v = 'the end'

            print('Trimming output data for {} from {} to {}.'.format(fd['meta']['name'], tmin_v, tmax_v))
        # Find time index range
        if tmin:
            T0ind = np.searchsorted(fd['Time'], tmin)
        else:
            T0ind = 0
        if tmax:
            Tfind = np.searchsorted(fd['Time'], tmax) + 1
        else: 
            Tfind = len(fd['Time'])

        if T0ind+1 > len(fd['Time']):
            raise ValueError('The initial time to trim {} to is after the end of the simulation.'.format(fd['meta']['name']))

        # # Modify time
        fd['Time'] = fd['Time'][T0ind:Tfind] - fd['Time'][T0ind]

        # Remove all vales in data where time is not in desired range
        for key in fd.keys():
            if key.lower() not in ['time', 'meta']:
                fd[key] = fd[key][T0ind:Tfind]


    return fast_data



import numpy as np
import os
# import pandas
from collections import OrderedDict

def turb_specs(V_ref, L_u, L_v, L_w, sigma_u, sigma_v, sigma_w, template_file, filename):
    
    f=(np.array([np.arange(0.0015873015873015873015873015873, 20.00001, 0.0015873015873015873015873015873)])).T
    
    a=int(len(f))
    
    U=np.zeros((a,1),dtype=float)
    V=np.zeros((a,1),dtype=float)
    W=np.zeros((a,1),dtype=float)
    
    for i in range(0,a):
        U[i,0]= (4*L_u/V_ref)*sigma_u**2/((1+6*f[i,0]*L_u/V_ref)**(5./3.))
        V[i,0]= (4*L_v/V_ref)*sigma_v**2/((1+6*f[i,0]*L_v/V_ref)**(5./3.))
        W[i,0]= (4*L_w/V_ref)*sigma_w**2/((1+6*f[i,0]*L_w/V_ref)**(5./3.))
        
    df=pandas.DataFrame({'Frequency (Hz)':f[:,0],'u-component PSD (m^2/s)': U[:,0],'v-component PSD (m^2/s)': V[:,0],'w-component PSD (m^2/s)':W[:,0]})
    with open(template_file, 'r') as f:
        get_all=f.readlines() 
    
    
        
    with open(filename,'w') as f:
        for i,line in enumerate(get_all):
                if i < 11:  ## STARTS THE NUMBERING FROM 1 (by default it begins with 0)
                        f.writelines(line)                             ## OVERWRITES line:2
                else: 
                        f.write(df.to_string(index=False,header=False,col_space=15))
                        break
                
                                                                                                               
    




import numpy as np
import os, sys, copy, itertools
import multiprocessing as mp

from ROSCO_toolbox.ofTools.case_gen.CaseGen_General import CaseGen_General, save_case_matrix, save_case_matrix_yaml
from ROSCO_toolbox.ofTools.fast_io.pyIECWind import pyIECWind_extreme, pyIECWind_turb

try:
    from mpi4py import MPI
except:
    pass
    
# Generate wind files
def gen_windfile(data):
    # function for calling wind file execution
    iecwind = data[0]
    IEC_WindType = data[1]
    change_vars = data[2]
    var_vals = data[3]

    if 'Seeds' in change_vars:
        iecwind.seed = var_vals[change_vars.index('Seeds')]
    U = var_vals[change_vars.index('U')]

    wind_file, wind_file_type = iecwind.execute(IEC_WindType, U)
    if type(wind_file) is str:
        U_out = [U]
        WindFile_out = [wind_file]
        WindFile_type_out = [wind_file_type]
    elif type(wind_file) is list:
        U_out = [U]*len(wind_file)
        WindFile_out = wind_file
        WindFile_type_out = wind_file_type
    return [U_out, WindFile_out, WindFile_type_out]

class CaseGen_IEC():

    def __init__(self):
        
        self.init_cond = {} # Dictionary of steady state operating conditions as a function of wind speed, used for setting inital conditions

        self.Turbine_Class               = 'I' # I, II, III, IV
        self.Turbulence_Class            = 'A'
        self.D                           = 126.
        self.z_hub                       = 90.

        # DLC inputs
        self.dlc_inputs = {}
        self.transient_dir_change        = 'both'  # '+','-','both': sign for transient events in EDC, EWS
        self.transient_shear_orientation = 'both'  # 'v','h','both': vertical or horizontal shear for EWS
        self.PC_MaxRat                   = 5.      # (deg), maximum blade pitch rate, used for emergency stop, DLC 5.1
        self.TMax                        = 0.
        self.TStart                      = 30.
        self.uniqueSeeds                 = False
        self.uniqueWaveSeeds             = False

        self.debug_level                 = 2
        self.parallel_windfile_gen       = False
        self.cores                       = 0
        self.overwrite                   = False

        self.mpi_run                     = False
        self.comm_map_down               = []

        self.run_dir          = os.path.dirname( os.path.dirname( os.path.dirname( os.path.realpath(__file__) ) ) ) + os.sep
        self.Turbsim_exe = os.path.join(self.run_dir, 'local/bin/turbsim')


    def execute(self, case_inputs={}):

        case_list_all = {}
        dlc_all = []

        for i, dlc in enumerate(self.dlc_inputs['DLC']):
            case_inputs_i = copy.deepcopy(case_inputs)

            # DLC specific variable changes
            if dlc in [1.1, 1.2, 5.1]:
                IEC_WindType = 'NTM'
                alpha = 0.2
                iecwind = pyIECWind_turb()
                if self.TMax == 0:
                    TMax = 720.
                else:
                    TMax = self.TMax

            elif dlc in [1.3, 6.1, 6.3]:
                if self.Turbine_Class == 'I':
                    x = 1
                elif self.Turbine_Class == 'II':
                    x = 2
                elif self.Turbine_Class == 'III':
                    x = 3
                else:
                    exit('Class of the WT is needed for the ETM wind, but it is currently not set to neither 1,2 or 3.')
                IEC_WindType = '%uETM'%x
                alpha = 0.11
                iecwind = pyIECWind_turb()
                if self.TMax == 0:
                    TMax = 720.
                else:
                    TMax = self.TMax

            elif dlc == 1.4:
                IEC_WindType = 'ECD'
                alpha = 0.2
                iecwind = pyIECWind_extreme()
                if self.TMax == 0:
                    TMax = 120.
                else:
                    TMax = self.TMax

            elif dlc == 1.5:
                IEC_WindType = 'EWS'
                alpha = 0.2
                iecwind = pyIECWind_extreme()
                if self.TMax == 0:
                    TMax = 120.
                else:
                    TMax = self.TMax

            # Windfile generation setup
            if self.TMax == 0.:
                iecwind.AnalysisTime = TMax
                iecwind.TF           = TMax
            else:
                iecwind.AnalysisTime = self.TMax
                iecwind.TF           = self.TMax

            iecwind.TStart           = self.TStart
            iecwind.Turbine_Class    = self.Turbine_Class
            iecwind.Turbulence_Class = self.Turbulence_Class
            iecwind.IEC_WindType     = IEC_WindType
            iecwind.dir_change       = self.transient_dir_change
            iecwind.shear_orient     = self.transient_shear_orientation
            iecwind.z_hub            = self.z_hub
            iecwind.D                = self.D
            iecwind.PLExp            = alpha
            
            iecwind.outdir           = self.wind_dir
            iecwind.case_name        = self.case_name_base
            iecwind.Turbsim_exe      = self.Turbsim_exe
            iecwind.debug_level      = self.debug_level
            iecwind.overwrite        = self.overwrite

            # Set DLC specific settings
            iecwind_ex = pyIECWind_extreme()
            iecwind_ex.Turbine_Class    = self.Turbine_Class
            iecwind_ex.Turbulence_Class = self.Turbulence_Class
            iecwind_ex.z_hub            = self.z_hub
            iecwind_ex.setup()
            _, V_e50, V_e1, V_50, V_1   = iecwind_ex.EWM(0.)


            if dlc == 5.1:
                case_inputs_i[("ServoDyn","TPitManS1")]  = {'vals':[self.TStart], 'group':0}
                case_inputs_i[("ServoDyn","TPitManS2")]  = {'vals':[self.TStart], 'group':0}
                case_inputs_i[("ServoDyn","TPitManS3")]  = {'vals':[self.TStart], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(1)")] = {'vals':[self.PC_MaxRat], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(2)")] = {'vals':[self.PC_MaxRat], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(3)")] = {'vals':[self.PC_MaxRat], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(1)")]  = {'vals':[90.], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(2)")]  = {'vals':[90.], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(3)")]  = {'vals':[90.], 'group':0}
                case_inputs_i[("ServoDyn","GenTiStp")]   = {'vals':["True"], 'group':0}
                case_inputs_i[("ServoDyn","TimGenOf")]   = {'vals':[self.TStart], 'group':0}
            else:
                case_inputs_i[("ServoDyn","TPitManS1")]  = {'vals':[9999.9], 'group':0}
                case_inputs_i[("ServoDyn","TPitManS2")]  = {'vals':[9999.9], 'group':0}
                case_inputs_i[("ServoDyn","TPitManS3")]  = {'vals':[9999.9], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(1)")] = {'vals':[2.], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(2)")] = {'vals':[2.], 'group':0}
                case_inputs_i[("ServoDyn","PitManRat(3)")] = {'vals':[2.], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(1)")]  = {'vals':[0.], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(2)")]  = {'vals':[0.], 'group':0}
                case_inputs_i[("ServoDyn","BlPitchF(3)")]  = {'vals':[0.], 'group':0}
                case_inputs_i[("ServoDyn","GenTiStp")]   = {'vals':["True"], 'group':0}
                case_inputs_i[("ServoDyn","TimGenOf")]   = {'vals':[9999.9], 'group':0}

            if dlc == 6.1:
                self.dlc_inputs['U'][i] = [V_50]
                self.dlc_inputs['Yaw'][i] = [-8.,8.]
                case_inputs_i[("ElastoDyn","GenDOF")]   = {'vals':["False"], 'group':0}
                case_inputs_i[("ElastoDyn","YawDOF")]   = {'vals':["False"], 'group':0}
                case_inputs_i[("ElastoDyn","RotSpeed")] = {'vals':[0.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch1")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch2")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch3")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ServoDyn","PCMode")]    = {'vals':[0], 'group':0}
                case_inputs_i[("AeroDyn15","AFAeroMod")]= {'vals':[1], 'group':0}
            elif dlc == 6.3:
                self.dlc_inputs['U'][i] = [V_1]
                self.dlc_inputs['Yaw'][i] = [-20.,20.]
                case_inputs_i[("ElastoDyn","GenDOF")]   = {'vals':["False"], 'group':0}
                case_inputs_i[("ElastoDyn","YawDOF")]   = {'vals':["False"], 'group':0}
                case_inputs_i[("ElastoDyn","RotSpeed")] = {'vals':[0.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch1")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch2")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ElastoDyn","BlPitch3")] = {'vals':[90.], 'group':0}
                case_inputs_i[("ServoDyn","PCMode")]    = {'vals':[0], 'group':0}
                case_inputs_i[("AeroDyn15","AFAeroMod")]= {'vals':[1], 'group':0}
            else:
                self.dlc_inputs['Yaw'][i] = [0.]
                case_inputs_i[("ServoDyn","PCMode")]    = {'vals':[5], 'group':0}
                case_inputs_i[("AeroDyn15","AFAeroMod")]= {'vals':[2], 'group':0}
                case_inputs_i[("ElastoDyn","GenDOF")]   = {'vals':["True"], 'group':0}


            # Matrix combining N dlc variables that affect wind file generation
            # Done so a single loop can be used for generating wind files in parallel instead of using nested loops
            var_list = ['U', 'Seeds']
            group_len = []
            change_vars = []
            change_vals = []
            for var in var_list:
                if len(self.dlc_inputs[var][i]) > 0:
                    group_len.append(len(self.dlc_inputs[var][i]))
                    change_vars.append(var)
                    change_vals.append(self.dlc_inputs[var][i])
            group_idx = [range(n) for n in group_len]
            matrix_idx = list(itertools.product(*group_idx))
            matrix_group_idx = [np.where([group_i == group_j for group_j in range(0,len(group_len))])[0].tolist() for group_i in range(0,len(group_len))]
            matrix_out = []
            for idx, row in enumerate(matrix_idx):
                row_out = [None]*len(change_vars)
                for j, val in enumerate(row):
                    for g in matrix_group_idx[j]:
                        row_out[g] = change_vals[g][val]
                matrix_out.append(row_out)
            matrix_out = np.asarray(matrix_out)

            # change seed manually to not mess up indexing above, 600
            if self.uniqueSeeds and dlc in [1.1, 1.2, 5.1, 1.3, 6.1, 6.3]:
                for idx, row in enumerate(matrix_out):
                    matrix_out[idx][1] = idx + 600
            
            if self.parallel_windfile_gen and not self.mpi_run:
                # Parallel wind file generation (threaded with multiprocessing)
                if self.cores != 0:
                    p = mp.Pool(self.cores)
                else:
                    p = mp.Pool()
                data_out = p.map(gen_windfile, [(iecwind, IEC_WindType, change_vars, var_vals) for var_vals in matrix_out])
                U_out = []
                WindFile_out = []
                WindFile_type_out = []
                for case in data_out:
                    U_out.extend(case[0])
                    WindFile_out.extend(case[1])
                    WindFile_type_out.extend(case[2])

            elif self.parallel_windfile_gen and self.mpi_run:
                # Parallel wind file generation with MPI
                comm = MPI.COMM_WORLD
                # size = comm.Get_size()
                rank = comm.Get_rank()
                sub_ranks = self.comm_map_down[rank]
                size = len(sub_ranks)

                N_cases = len(matrix_out)
                N_loops = int(np.ceil(float(N_cases)/float(size)))

                U_out = []
                WindFile_out = []
                WindFile_type_out = []
                for j in range(N_loops):
                    idx_s    = j*size
                    idx_e    = min((j+1)*size, N_cases)

                    for k, var_vals in enumerate(matrix_out[idx_s:idx_e]):
                        data   = [gen_windfile, [iecwind, IEC_WindType, change_vars, var_vals]]
                        rank_k = sub_ranks[k]
                        comm.send(data, dest=rank_k, tag=0)

                    for k, var_vals in enumerate(matrix_out[idx_s:idx_e]):
                        rank_k = sub_ranks[k]
                        data_out = comm.recv(source=rank_k, tag=1)
                        U_out.extend(data_out[0])
                        WindFile_out.extend(data_out[1])
                        WindFile_type_out.extend(data_out[2])

            else:
                # Serial
                U_out = []
                WindFile_out = []
                WindFile_type_out = []
                for var_vals in matrix_out:
                    [U_out_i, WindFile_out_i, WindFile_type_out_i] = gen_windfile([iecwind, IEC_WindType, change_vars, var_vals])
                    U_out.extend(U_out_i)
                    WindFile_out.extend(WindFile_out_i)
                    WindFile_type_out.extend(WindFile_type_out_i)
            
            # Set FAST variables from DLC setup
            if ("Fst","TMax") not in case_inputs_i:
                case_inputs_i[("Fst","TMax")] = {'vals':[TMax], 'group':0}
            case_inputs_i[("InflowWind","WindType")] = {'vals':WindFile_type_out, 'group':1}
            case_inputs_i[("InflowWind","Filename_Uni")] = {'vals':WindFile_out, 'group':1}
            case_inputs_i[("InflowWind","FileName_BTS")] = {'vals':WindFile_out, 'group':1}
            case_inputs_i[("InflowWind","RefLength")] = {'vals':[self.D], 'group':0}

            if len(self.dlc_inputs['Yaw'][i]) > 0:
                    case_inputs_i[("ElastoDyn","NacYaw")] = {'vals':self.dlc_inputs['Yaw'][i], 'group':2}

            # Set FAST variables from inital conditions
            if self.init_cond:
                for var in self.init_cond.keys():
                    if var not in case_inputs_i.keys():
                        inital_cond_i = [np.interp(U, self.init_cond[var]['U'], self.init_cond[var]['val']) for U in U_out]
                        case_inputs_i[var] = {'vals':inital_cond_i, 'group':1}
            
            # make unique wave seeds
            if self.uniqueWaveSeeds:
                seed_base   = int(float(dlc) * 10000)   # set wave seed based on dlc so no repeats
                num_in_dlc  = len(case_inputs_i[("InflowWind","FileName_BTS")]['vals'])  # sims in each DLC
                wave_seeds  = (seed_base + np.arange(0,num_in_dlc)).tolist()

                case_inputs_i[("HydroDyn","WaveSeed1")] = {'vals':wave_seeds, 'group':1}

            # Append current DLC to full list of cases
            case_list, case_name = CaseGen_General(case_inputs_i, self.run_dir, self.case_name_base, save_matrix=False)
            case_list_all = self.join_case_dicts(case_list_all, case_list)
            case_names_all = [self.case_name_base +'_'+ ('%d'%i).zfill(len('%d'%(len(case_list_all)-1))) for i in range(len(case_list_all))]
            dlc_all.extend([dlc]*len(case_list))

        # Save case matrix file
        self.save_joined_case_matrix(case_list_all, dlc_all, case_names_all)

        return case_list_all, case_names_all, dlc_all


    def join_case_dicts(self, caselist, caselist_add):
        if caselist:
            keys1 = caselist[0].keys()
            keys2 = caselist_add[0].keys()
            n1 = len(caselist)
            n2 = len(caselist_add)

            common = list(set(keys1) & set(keys2))
            missing1 = list(set(keys1).difference(keys2))
            missing2 = list(set(keys2).difference(keys1))

            # caselist_out = copy.copy(case_list)
            for i in range(n1):
                for var in missing2:
                    caselist[i][var] = np.nan
            for i in range(n2):
                for var in missing1:
                    caselist_add[i][var] = np.nan

            return caselist + caselist_add
        else:
            return caselist_add

    def save_joined_case_matrix(self, caselist, dlc_list, case_names_all):

        change_vars = sorted(caselist[0].keys())

        matrix_out = []
        for case in caselist:
            row_out = [None]*len(change_vars)
            for i, var in enumerate(change_vars):
                # row_out[i] = str(case[var])
                row_out[i] = case[var]
            matrix_out.append(row_out)

        change_vars = [('IEC', 'DLC')] + change_vars
        [matrix_row.insert(0, dlc) for dlc, matrix_row in zip(dlc_list,matrix_out)]

        if not os.path.exists(self.run_dir):
            os.makedirs(self.run_dir)

        save_case_matrix(matrix_out, change_vars, self.run_dir)
        save_case_matrix_yaml(matrix_out, change_vars, self.run_dir, case_names_all)



if __name__=="__main__":
    
    iec = CaseGen_IEC()

    # Turbine Data
    iec.init_cond = {} # can leave as {} if data not available
    iec.init_cond[("ElastoDyn","RotSpeed")] = {'U':[3., 4., 5., 6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25]}
    iec.init_cond[("ElastoDyn","RotSpeed")]['val'] = [6.972, 7.183, 7.506, 7.942, 8.469, 9.156, 10.296, 11.431, 11.89, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1]
    iec.init_cond[("ElastoDyn","BlPitch1")] = {'U':[3., 4., 5., 6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25]}
    iec.init_cond[("ElastoDyn","BlPitch1")]['val'] = [0., 0., 0., 0., 0., 0., 0., 0., 0., 3.823, 6.602, 8.668, 10.450, 12.055, 13.536, 14.920, 16.226, 17.473, 18.699, 19.941, 21.177, 22.347, 23.469]
    iec.init_cond[("ElastoDyn","BlPitch2")] = iec.init_cond[("ElastoDyn","BlPitch1")]
    iec.init_cond[("ElastoDyn","BlPitch3")] = iec.init_cond[("ElastoDyn","BlPitch1")]

    iec.Turbine_Class = 'I' # I, II, III, IV
    iec.Turbulence_Class = 'A'
    iec.D = 126.
    iec.z_hub = 90.

    # DLC inputs
    iec.dlc_inputs = {}
    iec.dlc_inputs['DLC']   = [1.1, 1.5]
    iec.dlc_inputs['U']     = [[8, 9, 10], [8]]
    iec.dlc_inputs['Seeds'] = [[5, 6, 7], []]
    iec.dlc_inputs['Yaw']   = [[], []]

    iec.transient_dir_change        = 'both'  # '+','-','both': sign for transient events in EDC, EWS
    iec.transient_shear_orientation = 'both'  # 'v','h','both': vertical or horizontal shear for EWS

    # Naming, file management, etc
    iec.wind_dir = 'temp/wind'
    iec.case_name_base = 'testing'
    iec.Turbsim_exe = 'C:/Users/egaertne/WT_Codes/Turbsim_v2.00.07/bin/TurbSim_x64.exe'
    iec.debug_level = 1
    iec.run_dir = 'temp'

    iec.parallel_windfile_gen = True
    iec.cores = 4

    # Run
    iec.execute()

    




'''
Run ROSCO lite testing scripts for controller functionality verification
'''
import os
import glob
import ROSCO_testing
import importlib

os.system("taskset -p 0xffffffffffff %d" % os.getpid())


def run_testing(turbine2test, testtype, rosco_binaries=[], discon_files=[], **kwargs):
    '''
    Run and post-process controller testing scripts. 

    NOTE: There are 

    Parameters:
    -----------
    turbine2test: str
        Reference turbine to test: IEA-15MW or NREL-5MW
    testtype: str
        Type of test: 'lite' - small set of 5 minute simulations
                      'heavy' - large set of 10 minute simulation (DLC 1.3 & 1.4 subset)
                      'binary-comp' - compare binaries
                      'discon-comp' - compare discon files
    rosco_binaries: list-like
        List of strings with paths to ROSCO binaries. Len>2 needed for binary-comp!
        NOTE: Will use the first binary if testtype='lite' or 'heavy'. s
    discon_files: list-like, only for discon-comp
        List of strings with paths to DISCON.IN files - Neede for discon-comp!
    ''' 
    ## =================== INITIALIZATION ===================
    # Instantiate  class
    rt = ROSCO_testing.ROSCO_testing(**kwargs)

    # Post Processing Parameters
    reCrunch = True                     # re-run pCrunch?

    #### ============================================ ####
    #                     THE ACTION                     #
    # ----- Shouldn't need to change anything here ----- #

    # Check validity inputs
    if testtype.lower() == 'binary-comp' and len(rosco_binaries) < 2:
        raise ValueError('You need at least two binaries to compare for testtype = "binary-comp"!')
    if testtype.lower() == 'discon-comp' and len(discon_files) < 2:
        raise ValueError('You need at least two discon files to compare for testtype = "binary-comp"!')

    # Setup test turbine
    if turbine2test == 'NREL-5MW':
        rt.Turbine_Class = 'I'
        rt.Turbulence_Class = 'A'
        rt.FAST_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../Test_Cases/NREL-5MW')
        rt.FAST_InputFile = 'NREL-5MW.fst'
    elif turbine2test == 'IEA-15MW':
        rt.Turbine_Class = 'I'
        rt.Turbulence_Class = 'B'
        rt.FAST_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../Test_Cases/IEA-15-240-RWT-UMaineSemi')
        rt.FAST_InputFile = 'IEA-15-240-RWT-UMaineSemi.fst'
    else:
        raise ValueError('{} is not an available turbine to test!'.format(turbine2test))

    # Run test
    if testtype.lower() == 'lite':
        rt.rosco_path = rosco_binaries[0]
        rt.ROSCO_Test_lite()
    elif testtype.lower() == 'heavy':
        rt.rosco_path = rosco_binaries[0]
        rt.ROSCO_Test_heavy()
    elif testtype.lower() == 'binary-comp':
        rt.ROSCO_Controller_Comp(rosco_binaries)
    elif testtype.lower() == 'discon-comp':
        rt.ROSCO_DISCON_Comp(discon_files)
    else:
        raise ValueError('{} is an invalid test type!'.format(testtype))


        
if __name__ == "__main__":

    # WEIS directory, for running openfast, etc.
    this_dir =  os.path.dirname(os.path.realpath(__file__))

    # Setup ROSCO testing parameters
    rt_kwargs = {} 
    rt_kwargs['runDir']     = os.path.join('/scratch/dzalkind/ROSCO_testing','version2.3.0')        # directory for FAST simulations
    rt_kwargs['namebase']   = 'heavy_test'     # Base name for FAST files 
    rt_kwargs['FAST_exe']   = '/home/dzalkind/Tools/openfast-main/install/bin/openfast'       # OpenFAST executable path
    rt_kwargs['wind_dir']   = os.path.join('/scratch/dzalkind/ROSCO_testing','wind','IEA-15_heavy')       # OpenFAST executable path
    rt_kwargs['Turbsim_exe']= 'turbsim'    # Turbsim executable path
    rt_kwargs['FAST_ver']   = 'OpenFAST'            # FAST version
    rt_kwargs['dev_branch'] = True                  # dev branch of Openfast?
    rt_kwargs['debug_level']= 2                     # debug level. 0 - no outputs, 1 - minimal outputs, 2 - all outputs
    rt_kwargs['overwrite']  = False                 # overwite fast sims?
    rt_kwargs['cores']      = 36                     # number of cores if multiprocessing
    rt_kwargs['mpi_run']    = False                 # run using mpi
    rt_kwargs['mpi_comm_map_down'] = []             # core mapping for MPI
    rt_kwargs['outfile_fmt'] = 2                    # 1 = .txt, 2 = binary, 3 = both
    # rt_kwargs['comp_dir']   = '/projects/ssc/dzalkind/ROSCO/ROSCO_testing/version2.2.0'                    # 1 = .txt, 2 = binary, 3 = both
    rt_kwargs['comp_dir']   = '/scratch/dzalkind/ROSCO_testing/pr_46'                    # 1 = .txt, 2 = binary, 3 = both
    

    # ---- Define test type ----
    turbine2test = 'IEA-15MW'   # IEA-15MW or NREL-5MW
    testtype     = 'heavy'       # lite, heavy, binary-comp, discon-comp

    # Only fill one of these if comparing controllers
    rosco_binaries = [glob.glob(os.path.join(this_dir,'../ROSCO/build/libdiscon.*'))[0]] # Differently named libdiscons to compare
    discon_files = []   # Differently named DISCON.IN files to compare

    # Run testing
    run_testing(turbine2test, testtype, rosco_binaries=rosco_binaries, discon_files=discon_files, **rt_kwargs)
